#include "stdafx.h"
#include "critical_params.h"

#ifndef __CELL_H__
#define __CELL_H__

#define MFP_UNIT	4

enum CellTypes { CELL_NONE = 0, CELL_ALL = 0, CELL_H = 1, CELL_K = 2, CELL_D = 3,
	CELL_TREG = 4, //CELL_TG = 4, 
	CELL_CD4 = 5, CELL_CD8 = 6, CELL_M = 7,
	INPUT_HEV = 8, INPUT_AFFERENT = 9,
	OUTPUT_EFFERENT = 10,
	GRID_NOGO = 11, GRID_FRC = 12, NUMCELLTYPES = 13
};

enum CytokineTypes { CK_TNF = 0, CK_CCL5, CK_CXCL9, CK_CCL2, CK_TOTAL};

enum CellStates {
	CELL_DEAD = 0,// CELL_COMPLETE = 0,

	// LN
	HCELL_DEAD = 0, HCELL_RESTING = 1, HCELL_BOUND = 2,
	HCELL_ACTIVE = 3, HCELL_EFFECTOR = 4, HCELL_BOUND_TO_IDC = 5,
//HCELL_NOTMOVE_BY_DC KCELL_NOTMOVE_BY_LD
	KCELL_DEAD = 0, KCELL_RESTING = 1, KCELL_BOUND = 2,
	KCELL_ACTIVE = 3, KCELL_EFFECTOR = 4, KCELL_MEMORY = 5,

	DCELL_DEAD = 0, DCELL_IMMATURE = 1, DCELL_MATURE = 2, DCELL_LICENSED = 3,

	// GR
	//TGCELL_DEAD = 0, TGCELL_RESTING = 1, TGCELL_ACTIVE = 2, TGCELL_REG = 3,
	TREGCELL_DEAD = 0, TREGCELL_RESTING = 1, TREGCELL_ACTIVE = 2, TREGCELL_REG = 3,

	//	TCCELL_DEAD = 0, TCCELL_RESTING = 1, TCCELL_ACTIVE = 2,

	//	TRCELL_DEAD = 0, TRCELL_RESTING = 1, TRCELL_ACTIVE = 2,
	CD4CELL_DEAD = 0, CD4CELL_RESTING = 1, CD4CELL_ACTIVE = 2,
	CD8CELL_DEAD = 0, CD8CELL_RESTING = 1, CD8CELL_ACTIVE = 2,

	MCELL_DEAD = 0, MCELL_RESTING = 1, MCELL_INFECTED = 2, MCELL_CINFECTED = 3,
	MCELL_ACTIVE = 4
};

#ifdef WIN32
#pragma pack(push, 1)
#endif

class Cell {
	public:
#ifdef WIN32
		// one year = 5256000 ticks
		u32 enter_time:26; // 0-64mil
		u32 state:3; // 0-7
		u32 next_state:3; // 0-7

		s32 birth_time:27; // +-64mil
		u32 lastdir:5; // 0-31

		u32 death_time:26; // 0-64mil
		u32 celltype:4; // 0-15
		u32 displaybig:1; // binary
		u32 movedyet:1; // binary
#else
		CellTypes celltype:4; // 0-15
		bool displaybig:1; // binary
		bool movedyet:1; // binary
		int lastdir:5; // 0-31
		CellStates state:3, next_state:3; // 0-7
		// 14 bits
		int :2; // padding

		// one year = 5256000 ticks
		int enter_time:24; // 0-16mil
		int birth_time:24;
		int death_time:24;
#endif
}; // 11 bytes

class LNCell : public Cell {
	public:
		int mfp;

		// T cells
		bool cognate;
//		bool dc_bound;
		int dc_bound_time;
		int dc_unbind_time;
		float dc_unbind_pMHC;
		coord dc_coord;
		int division_counter;
		float priming_history;
		int hide_time;

		// D cells
		bool lockdown;
		float pMHC;
//		int license_time;
};

class GRCell : public Cell {
	public:
/*		// T cells
		bool cognate;
		int dc_bound_time;
		int dc_unbind_time;
		float dc_unbind_pMHC;
		coord dc_coord;
		int division_counter;
		float priming_history;
		int hide_time;

		// D cells
		bool lockdown;
		float pMHC;
//		int license_time;*/
};

#ifdef WIN32
#pragma pack(pop)
#endif

#endif
