/*
Granuloma model notes

Grid size:
//old = 20 um, 100x100x100
-> 4um, 400x400x400

Initial grid:
Sphere packing scoops for empty spaces, size 200-300 um = 250 um + 50 um tissue area (total)
= diameter 60 elements -> 52%, tissue = diameter 6 elements on each side
20% vascularization of lung
Empty 65%
15% tissue, 15% blood, 5% lymphatic
Open, Tissue, Blood, Lymphatic
2001 macrophages - 2000 resting, one infected
200 dendritic cells - immature

Movement:
Macrophage & Dendritic cells, one iteration per 240 seconds = 40 ticks
T cells, one iteration per 120 seconds = 20 ticks
Force equation to encourage movement away from sources, allow cells through
Build into empty space, need to be next to other cells??
Tissue spaces can be moved into by cells
Blood spaces cannot be moved into
Lymphatic spaces cause cell exit if mature DC, else cannot move there
If tries to go off-grid, let them

Diffusion:
 Res: 100x100x100 - easily adjustable - 200x200x200??
 Step: 6 seconds @ 20um (max 10)
 No diffusion in air, only in tissue
CCL2 - float (typedef so we can change)
CCL5 - float
CXCL9 - float
TNF - float
BE - float

Recruitment:
Some sources are enabled or disabled based on chemokines
Macrophages come out into open space

Drawing:
Show only blood vessels of certain size and certain % of smaller
Need to generate blood vessels of variable thickness

*/

#include "stdafx.h"
#include "VizGrid.h"

bool ENABLE_GROUTLINE = false;

void GRGrid::InitStructure() {
//Empty ~65%
//15% tissue, 15% blood, 5% lymphatic = 35%; 0.43, 0.43, 0.14
// ELEMENT_OPEN, ELEMENT_TISSUE, ELEMENT_BLOOD, ELEMENT_LYMPHATIC

	// default spaces to random type, default cells to empty
//	u32 inittime = clock();
	for (int x = 0; x < GR_GRID_X; x++) {
		for (int y = 0; y < GR_GRID_Y; y++) {
			for (int z = 0; z < GR_GRID_Z; z++) {
				float decision = rnd.getfloat();
				if (decision < 0.43)
					grid[x][y][z].elementtype = ELEMENT_TISSUE;
				else if (decision < 0.86)
					grid[x][y][z].elementtype = ELEMENT_BLOOD;
				else
					grid[x][y][z].elementtype = ELEMENT_LYMPHATIC;

				grid[x][y][z].cell.birth_time = 0;
				grid[x][y][z].cell.celltype = CELL_NONE;
				grid[x][y][z].cell.death_time = 0;
				grid[x][y][z].cell.displaybig = false;
				grid[x][y][z].cell.enter_time = 0;
				grid[x][y][z].cell.lastdir = 0;
				grid[x][y][z].cell.next_state = CELL_DEAD;
				grid[x][y][z].cell.state = CELL_DEAD;
			}
		}
	}
//	printf("Init time: %0.3fs\n", (float)(clock() - inittime)/(float)CLOCKS_PER_SEC);

	// now scoop out spheres to represent open space
	int numcleared = 0;
	for (int i = 0; i < 6; i++) {
		for (int j = 0; j < 6; j++) {
			for (int k = 0; k < 6; k++) {
				coordc curr(20 + i * 72, 20 + j * 72, 20 + k * 72);
				numcleared += ClearSphere(curr, 30); // = diameter 60
			}
		}
	}
	printf("Open space generated, %d/%d = %0.2f%%\n", numcleared, GR_GRID_X*GR_GRID_Y*GR_GRID_Z, 100.0f * (float)numcleared/(float)(GR_GRID_X*GR_GRID_Y*GR_GRID_Z));
	// nets 35%, not 65%

	// create initial cell population
	// put infected macrophage in middle-ish
	coordc inf = GenEmptyCoord(coordc(200, 200, 200), 50); // approximately middle
	grid[inf.x][inf.y][inf.z].cell.birth_time = 0;
	grid[inf.x][inf.y][inf.z].cell.celltype = CELL_M;
	grid[inf.x][inf.y][inf.z].cell.death_time = DC_MAX_AGE;
	grid[inf.x][inf.y][inf.z].cell.displaybig = 1;
	grid[inf.x][inf.y][inf.z].cell.enter_time = 0;
	grid[inf.x][inf.y][inf.z].cell.lastdir = rnd.get_inrange(0, 25);
	grid[inf.x][inf.y][inf.z].cell.movedyet = false;
	grid[inf.x][inf.y][inf.z].cell.next_state = MCELL_INFECTED;
	grid[inf.x][inf.y][inf.z].cell.state = MCELL_INFECTED;

	for (int j = 0; j < 2000; j++) {
		inf = GenEmptyCoord(coordc(200, 200, 200), -1); // everywhere
		grid[inf.x][inf.y][inf.z].cell.birth_time = 0;
		grid[inf.x][inf.y][inf.z].cell.celltype = CELL_M;
		grid[inf.x][inf.y][inf.z].cell.death_time = DC_MAX_AGE;
		grid[inf.x][inf.y][inf.z].cell.displaybig = 1;
		grid[inf.x][inf.y][inf.z].cell.enter_time = 0;
		grid[inf.x][inf.y][inf.z].cell.lastdir = rnd.get_inrange(0, 25);
		grid[inf.x][inf.y][inf.z].cell.movedyet = false;
		grid[inf.x][inf.y][inf.z].cell.next_state = MCELL_RESTING;
		grid[inf.x][inf.y][inf.z].cell.state = MCELL_RESTING;
	}

	for (int j = 0; j < 200; j++) {
		inf = GenEmptyCoord(coordc(200, 200, 200), -1); // everywhere
		grid[inf.x][inf.y][inf.z].cell.birth_time = 0;
		grid[inf.x][inf.y][inf.z].cell.celltype = CELL_D;
		grid[inf.x][inf.y][inf.z].cell.death_time = DC_MAX_AGE;
		grid[inf.x][inf.y][inf.z].cell.displaybig = 1;
		grid[inf.x][inf.y][inf.z].cell.enter_time = 0;
		grid[inf.x][inf.y][inf.z].cell.lastdir = rnd.get_inrange(0, 25);
		grid[inf.x][inf.y][inf.z].cell.movedyet = false;
		grid[inf.x][inf.y][inf.z].cell.next_state = DCELL_IMMATURE;
		grid[inf.x][inf.y][inf.z].cell.state = DCELL_IMMATURE;
	}

	for (int x = 0; x < GR_DIF_X; x++) {
		for (int y = 0; y < GR_DIF_Y; y++) {
			for (int z = 0; z < GR_DIF_Z; z++) {
				for (int p = 0; p < NUM_CHEMTYPES; p++) {
					chem[x][y][z].curr[p] = 0.0f;
					chem[x][y][z].next[p] = 0.0f;
				}
			}
		}
	}
	
	// for debugging cell distribution
/*		char name[20];
		sprintf(name, "%d.csv", 0);
		FILE * outf = fopen(name, "w");
		if (outf) {
			for (int x = 0; x < GR_GRID_X; x++)
				for (int y = 0; y < GR_GRID_Y; y++)
					for (int z = 0; z < GR_GRID_Z; z++)
						if (grid[x][y][z].cell.celltype != CELL_NONE)
							fprintf(outf, "%03d,%03d,%03d\n", x, y, z);
			fclose(outf);
		} else
			printf("ERROR OPENING FILE %s\n", name);
			*/
}

bool GRGrid::CalcTimeSlice(int slice) { // now 6 seconds per slice, 3 per 18s // 6 seconds per slice, 20 per 2 minutes
	timeindex = slice;
	if (slice % DIFFUSION_TIMESTEP == 0)
		CalcDiffusion(DIFFUSION_TIMESTEP);

	if (slice % 3 == 0) { // T cells
		ClearMoved();
		Recruitment();
		Movement();
		States();
		ClearDead();

		if (slice % 60 == 0) { // every 6 mins
			printf(" %0.1fh: [TREG] %d, [CD4] %d, [CD8] %d, [D] %d, [M] %d\n", (float)timeindex/600.0, totalcells[CELL_TREG], totalcells[CELL_CD4], totalcells[CELL_CD8], totalcells[CELL_D], totalcells[CELL_M]);
			if (totalcellslost > 0)
				printf("  Avg grid time: %0.1fh\n", (float)totalgridtime/(float)(totalcellslost * 600));
	//		if (nummfp > 0)
	//			printf("  Mean Free Path: %0.1fum\n", mfpsum/nummfp);
		}
	
		return true;
	}
	return false;
}

void GRGrid::CalcDiffusion(int dt) {
	//ClearDiffusionEdges(); // allows chemokines to disappear off grid (not needed as we aren't updating the edges anyhow)

	// NEW DIFFUSION
	// 0.10 min/min w/o diffusion
	// 0.17 min/min w/ diffusion @ /2
	// 0.12 min/min w/ diffusion @ /4
	// 0.12 min/min w/ diffusion @ /8

	// NEW DIFFUSION w/ INTERPOLATION
	// 0.49 min/min w/ diffusion @ /2

	// OLD DIFFUSION
	// 0.10 min/min w/o diffusion
	// 4.79 min/min w/ diffusion @ /2
	// 0.20 min/min w/ diffusion @ /4
	// 0.11 min/min w/ diffusion @ /8

	for (int x = 1; x < GR_DIF_X - 1; x++) {
		for (int y = 1; y < GR_DIF_Y - 1; y++) {
			for (int z = 1; z < GR_DIF_Z - 1; z++) {
				chem[x][y][z].next[CHEM_TNF] = (1 - D_TNF) * chem[x][y][z].curr[CHEM_TNF] + 
					( chem[x-1][y][z].curr[CHEM_TNF] + chem[x+1][y][z].curr[CHEM_TNF] + 
						chem[x][y+1][z].curr[CHEM_TNF] + chem[x][y-1][z].curr[CHEM_TNF] + 
						chem[x][y][z+1].curr[CHEM_TNF] + chem[x][y][z-1].curr[CHEM_TNF]
					) * (D_TNF / 6.0f) - deg_TNF * chem[x][y][z].curr[CHEM_TNF];

				chem[x][y][z].next[CHEM_CCL2] = (1 - D_CCL2) * chem[x][y][z].curr[CHEM_CCL2] + 
					( chem[x-1][y][z].curr[CHEM_CCL2] + chem[x+1][y][z].curr[CHEM_CCL2] + 
						chem[x][y+1][z].curr[CHEM_CCL2] + chem[x][y-1][z].curr[CHEM_CCL2] + 
						chem[x][y][z+1].curr[CHEM_CCL2] + chem[x][y][z-1].curr[CHEM_CCL2]
					) * (D_CCL2 / 6.0f) - deg_CCL2 * chem[x][y][z].curr[CHEM_CCL2];

				chem[x][y][z].next[CHEM_CCL5] = (1 - D_CCL5) * chem[x][y][z].curr[CHEM_CCL5] + 
					( chem[x-1][y][z].curr[CHEM_CCL5] + chem[x+1][y][z].curr[CHEM_CCL5] + 
						chem[x][y+1][z].curr[CHEM_CCL5] + chem[x][y-1][z].curr[CHEM_CCL5] + 
						chem[x][y][z+1].curr[CHEM_CCL5] + chem[x][y][z-1].curr[CHEM_CCL5]
					) * (D_CCL5 / 6.0f) - deg_CCL5 * chem[x][y][z].curr[CHEM_CCL5];

				chem[x][y][z].next[CHEM_CXCL9] = (1 - D_CXCL9) * chem[x][y][z].curr[CHEM_CXCL9] + 
					( chem[x-1][y][z].curr[CHEM_CXCL9] + chem[x+1][y][z].curr[CHEM_CXCL9] + 
						chem[x][y+1][z].curr[CHEM_CXCL9] + chem[x][y-1][z].curr[CHEM_CXCL9] + 
						chem[x][y][z+1].curr[CHEM_CXCL9] + chem[x][y][z-1].curr[CHEM_CXCL9]
					) * (D_CXCL9 / 6.0f) - deg_CXCL9 * chem[x][y][z].curr[CHEM_CXCL9];

				chem[x][y][z].next[CHEM_EXTMTB] = (1 - D_EXTMTB) * chem[x][y][z].curr[CHEM_EXTMTB] + 
					( chem[x-1][y][z].curr[CHEM_EXTMTB] + chem[x+1][y][z].curr[CHEM_EXTMTB] + 
						chem[x][y+1][z].curr[CHEM_EXTMTB] + chem[x][y-1][z].curr[CHEM_EXTMTB] + 
						chem[x][y][z+1].curr[CHEM_EXTMTB] + chem[x][y][z-1].curr[CHEM_EXTMTB]
					) * (D_EXTMTB / 6.0f) - deg_EXTMTB * chem[x][y][z].curr[CHEM_EXTMTB];

				chem[x][y][z].curr[CHEM_TNF] = chem[x][y][z].next[CHEM_TNF];
				chem[x][y][z].curr[CHEM_CCL2] = chem[x][y][z].next[CHEM_CCL2];
				chem[x][y][z].curr[CHEM_CCL5] = chem[x][y][z].next[CHEM_CCL5];
				chem[x][y][z].curr[CHEM_CXCL9] = chem[x][y][z].next[CHEM_CXCL9];
				chem[x][y][z].curr[CHEM_EXTMTB] = chem[x][y][z].next[CHEM_EXTMTB];
			}
		}
	}
}

// currently only handles decimation by 2
#if (DECIMATE_DIFFUSION != 2)
#error Currently requires diffusion decimation to be 2
#endif
DIFFUSETYPE GRGrid::bilinear_3d(coordc currloc, ChemTypes toprocess) {
/*	For decimation by 2:
z=0	E	2	E	2	E // if 2 of 3 axes are mod 2, go +- last axis and average (2 points)
	2	1	2	1	2 // if 1 of 3 axes are mod 2, go +- 2 axes and average (4 points)
	E	2	E	2	E

z=1	2	1	2	1	2
	1	0	1	0	1 // if 0 of 3 axes are mod 2, go +- 3 axes and average (8 points)
	2	1	2	1	2

z=2	E	2	E	2	E
	2	1	2	1	2
	E	2	E	2	E
*/
	u8 xm2 = 0, ym2 = 0, zm2 = 0;
	if ((currloc.x % DECIMATE_DIFFUSION) == 0)
		xm2 = 1;
	if ((currloc.y % DECIMATE_DIFFUSION) == 0)
		ym2 = 1;
	if ((currloc.z % DECIMATE_DIFFUSION) == 0)
		zm2 = 1;

	u8 axes = xm2 + ym2 + zm2;
	if (axes == 3) { // use direct coordinates
		return chem[currloc.x/DECIMATE_DIFFUSION][currloc.y/DECIMATE_DIFFUSION][currloc.z/DECIMATE_DIFFUSION].curr[toprocess];
	} else if (axes == 2) { // use 2-point interpolation
		if (xm2 == 0) { // +- x
			DIFFUSETYPE f1 = chem[(currloc.x - 1)/DECIMATE_DIFFUSION][currloc.y/DECIMATE_DIFFUSION][currloc.z/DECIMATE_DIFFUSION].curr[toprocess],
				f2 = chem[(currloc.x + 1)/DECIMATE_DIFFUSION][currloc.y/DECIMATE_DIFFUSION][currloc.z/DECIMATE_DIFFUSION].curr[toprocess];
			return (f1 + f2) / 2.0f;
		} else if (ym2 == 0) { // +- y
			DIFFUSETYPE f1 = chem[currloc.x/DECIMATE_DIFFUSION][(currloc.y - 1)/DECIMATE_DIFFUSION][currloc.z/DECIMATE_DIFFUSION].curr[toprocess],
				f2 = chem[currloc.x/DECIMATE_DIFFUSION][(currloc.y + 1)/DECIMATE_DIFFUSION][currloc.z/DECIMATE_DIFFUSION].curr[toprocess];
			return (f1 + f2) / 2.0f;
		} else if (zm2 == 0) { // +- z
			DIFFUSETYPE f1 = chem[currloc.x/DECIMATE_DIFFUSION][currloc.y/DECIMATE_DIFFUSION][(currloc.z - 1)/DECIMATE_DIFFUSION].curr[toprocess],
				f2 = chem[currloc.x/DECIMATE_DIFFUSION][currloc.y/DECIMATE_DIFFUSION][(currloc.z + 1)/DECIMATE_DIFFUSION].curr[toprocess];
			return (f1 + f2) / 2.0f;
		}
	} else if (axes == 1) { // use 4-point interpolation
		if (xm2 == 1) { // +- y, z
			DIFFUSETYPE f1 = chem[currloc.x/DECIMATE_DIFFUSION][(currloc.y - 1)/DECIMATE_DIFFUSION][(currloc.z - 1)/DECIMATE_DIFFUSION].curr[toprocess],
				f2 = chem[currloc.x/DECIMATE_DIFFUSION][(currloc.y - 1)/DECIMATE_DIFFUSION][(currloc.z + 1)/DECIMATE_DIFFUSION].curr[toprocess],
				f3 = chem[currloc.x/DECIMATE_DIFFUSION][(currloc.y + 1)/DECIMATE_DIFFUSION][(currloc.z - 1)/DECIMATE_DIFFUSION].curr[toprocess],
				f4 = chem[currloc.x/DECIMATE_DIFFUSION][(currloc.y + 1)/DECIMATE_DIFFUSION][(currloc.z + 1)/DECIMATE_DIFFUSION].curr[toprocess];
			return (f1 + f2 + f3 + f4) / 4.0f;
		} else if (ym2 == 1) { // +- x, z
			DIFFUSETYPE f1 = chem[(currloc.x - 1)/DECIMATE_DIFFUSION][currloc.y/DECIMATE_DIFFUSION][(currloc.z - 1)/DECIMATE_DIFFUSION].curr[toprocess],
				f2 = chem[(currloc.x - 1)/DECIMATE_DIFFUSION][currloc.y/DECIMATE_DIFFUSION][(currloc.z + 1)/DECIMATE_DIFFUSION].curr[toprocess],
				f3 = chem[(currloc.x + 1)/DECIMATE_DIFFUSION][currloc.y/DECIMATE_DIFFUSION][(currloc.z - 1)/DECIMATE_DIFFUSION].curr[toprocess],
				f4 = chem[(currloc.x + 1)/DECIMATE_DIFFUSION][currloc.y/DECIMATE_DIFFUSION][(currloc.z + 1)/DECIMATE_DIFFUSION].curr[toprocess];
			return (f1 + f2 + f3 + f4) / 4.0f;
		} else if (zm2 == 1) { // +- x, y
			DIFFUSETYPE f1 = chem[(currloc.x - 1)/DECIMATE_DIFFUSION][(currloc.y - 1)/DECIMATE_DIFFUSION][currloc.z/DECIMATE_DIFFUSION].curr[toprocess],
				f2 = chem[(currloc.x - 1)/DECIMATE_DIFFUSION][(currloc.y + 1)/DECIMATE_DIFFUSION][currloc.z/DECIMATE_DIFFUSION].curr[toprocess],
				f3 = chem[(currloc.x + 1)/DECIMATE_DIFFUSION][(currloc.y - 1)/DECIMATE_DIFFUSION][currloc.z/DECIMATE_DIFFUSION].curr[toprocess],
				f4 = chem[(currloc.x + 1)/DECIMATE_DIFFUSION][(currloc.y + 1)/DECIMATE_DIFFUSION][currloc.z/DECIMATE_DIFFUSION].curr[toprocess];
			return (f1 + f2 + f3 + f4) / 4.0f;
		}
	} else if (axes == 0) { // use 8-point interpolation
		DIFFUSETYPE f1 = chem[(currloc.x - 1)/DECIMATE_DIFFUSION][(currloc.y - 1)/DECIMATE_DIFFUSION][(currloc.z - 1)/DECIMATE_DIFFUSION].curr[toprocess],
			f2 = chem[(currloc.x - 1)/DECIMATE_DIFFUSION][(currloc.y - 1)/DECIMATE_DIFFUSION][(currloc.z + 1)/DECIMATE_DIFFUSION].curr[toprocess],
			f3 = chem[(currloc.x - 1)/DECIMATE_DIFFUSION][(currloc.y + 1)/DECIMATE_DIFFUSION][(currloc.z - 1)/DECIMATE_DIFFUSION].curr[toprocess],
			f4 = chem[(currloc.x - 1)/DECIMATE_DIFFUSION][(currloc.y + 1)/DECIMATE_DIFFUSION][(currloc.z + 1)/DECIMATE_DIFFUSION].curr[toprocess],
			f5 = chem[(currloc.x + 1)/DECIMATE_DIFFUSION][(currloc.y - 1)/DECIMATE_DIFFUSION][(currloc.z - 1)/DECIMATE_DIFFUSION].curr[toprocess],
			f6 = chem[(currloc.x + 1)/DECIMATE_DIFFUSION][(currloc.y - 1)/DECIMATE_DIFFUSION][(currloc.z + 1)/DECIMATE_DIFFUSION].curr[toprocess],
			f7 = chem[(currloc.x + 1)/DECIMATE_DIFFUSION][(currloc.y + 1)/DECIMATE_DIFFUSION][(currloc.z - 1)/DECIMATE_DIFFUSION].curr[toprocess],
			f8 = chem[(currloc.x + 1)/DECIMATE_DIFFUSION][(currloc.y + 1)/DECIMATE_DIFFUSION][(currloc.z + 1)/DECIMATE_DIFFUSION].curr[toprocess];
		return (f1 + f2 + f3 + f4 + f5 + f6 + f7 + f8) / 8.0f;
	}

	// if no return complain
	printf("ERROR!  Missing case in interpolation: %d %d %d\n", currloc.x, currloc.y, currloc.z);
	return 0.0f;
}

void GRGrid::DrawGrid() {
//	static float direction = 0.0f;
//	int dir = (int)floor(direction) % 26;

	glPushMatrix();

	glDisable(GL_LIGHTING);
#ifdef DISPGRIDBOX
	glColor3f(0.5, 0.5, 0.5);
	glBegin(GL_LINES);
	glVertex3f((float)GR_GRID_X/2.0-0.5, (float)GR_GRID_Y/2.0-0.5, (float)GR_GRID_Z/2.0-0.5);
	glVertex3f(-(float)GR_GRID_X/2.0-0.5, (float)GR_GRID_Y/2.0-0.5, (float)GR_GRID_Z/2.0-0.5);
	glVertex3f((float)GR_GRID_X/2.0-0.5, -(float)GR_GRID_Y/2.0-0.5, (float)GR_GRID_Z/2.0-0.5);
	glVertex3f(-(float)GR_GRID_X/2.0-0.5, -(float)GR_GRID_Y/2.0-0.5, (float)GR_GRID_Z/2.0-0.5);
	glVertex3f((float)GR_GRID_X/2.0-0.5, (float)GR_GRID_Y/2.0-0.5, (float)GR_GRID_Z/2.0-0.5);
	glVertex3f((float)GR_GRID_X/2.0-0.5, -(float)GR_GRID_Y/2.0-0.5, (float)GR_GRID_Z/2.0-0.5);
	glVertex3f(-(float)GR_GRID_X/2.0-0.5, (float)GR_GRID_Y/2.0-0.5, (float)GR_GRID_Z/2.0-0.5);
	glVertex3f(-(float)GR_GRID_X/2.0-0.5, -(float)GR_GRID_Y/2.0-0.5, (float)GR_GRID_Z/2.0-0.5);

	glVertex3f((float)GR_GRID_X/2.0-0.5, (float)GR_GRID_Y/2.0-0.5, -(float)GR_GRID_Z/2.0-0.5);
	glVertex3f(-(float)GR_GRID_X/2.0-0.5, (float)GR_GRID_Y/2.0-0.5, -(float)GR_GRID_Z/2.0-0.5);
	glVertex3f((float)GR_GRID_X/2.0-0.5, -(float)GR_GRID_Y/2.0-0.5, -(float)GR_GRID_Z/2.0-0.5);
	glVertex3f(-(float)GR_GRID_X/2.0-0.5, -(float)GR_GRID_Y/2.0-0.5, -(float)GR_GRID_Z/2.0-0.5);
	glVertex3f((float)GR_GRID_X/2.0-0.5, (float)GR_GRID_Y/2.0-0.5, -(float)GR_GRID_Z/2.0-0.5);
	glVertex3f((float)GR_GRID_X/2.0-0.5, -(float)GR_GRID_Y/2.0-0.5, -(float)GR_GRID_Z/2.0-0.5);
	glVertex3f(-(float)GR_GRID_X/2.0-0.5, (float)GR_GRID_Y/2.0-0.5, -(float)GR_GRID_Z/2.0-0.5);
	glVertex3f(-(float)GR_GRID_X/2.0-0.5, -(float)GR_GRID_Y/2.0-0.5, -(float)GR_GRID_Z/2.0-0.5);

	glVertex3f((float)GR_GRID_X/2.0-0.5, (float)GR_GRID_Y/2.0-0.5, (float)GR_GRID_Z/2.0-0.5);
	glVertex3f((float)GR_GRID_X/2.0-0.5, (float)GR_GRID_Y/2.0-0.5, -(float)GR_GRID_Z/2.0-0.5);
	glVertex3f((float)GR_GRID_X/2.0-0.5, -(float)GR_GRID_Y/2.0-0.5, (float)GR_GRID_Z/2.0-0.5);
	glVertex3f((float)GR_GRID_X/2.0-0.5, -(float)GR_GRID_Y/2.0-0.5, -(float)GR_GRID_Z/2.0-0.5);
	glVertex3f(-(float)GR_GRID_X/2.0-0.5, (float)GR_GRID_Y/2.0-0.5, (float)GR_GRID_Z/2.0-0.5);
	glVertex3f(-(float)GR_GRID_X/2.0-0.5, (float)GR_GRID_Y/2.0-0.5, -(float)GR_GRID_Z/2.0-0.5);
	glVertex3f(-(float)GR_GRID_X/2.0-0.5, -(float)GR_GRID_Y/2.0-0.5, (float)GR_GRID_Z/2.0-0.5);
	glVertex3f(-(float)GR_GRID_X/2.0-0.5, -(float)GR_GRID_Y/2.0-0.5, -(float)GR_GRID_Z/2.0-0.5);
	glEnd();
#endif

	if (ENABLE_GROUTLINE) {
		glColor3f(0.3, 0.3, 0.3);
		// now draw new grid boundaries
		glBegin(GL_LINES);
#ifdef DISPGRIDZ
		// iterate through x with y > GR_GRID_Y/2 and y < GR_GRID_Y/2 and find min y
		for (int z = 0; z < GR_GRID_Z; z++) {
			for (int x = 0; x < GR_GRID_X; x++) {
				int y1 = GR_GRID_Y - 1;
				for (int y = GR_GRID_Y/2; y < GR_GRID_Y; y++) {
					if (grid[x][y][z].cell.celltype == GRID_NOGO) {
						y1 = y;
						break;
					}
				}
				if (x != 0)
					glVertex3f(x-GR_GRID_X/2,y1-GR_GRID_Y/2,z-GR_GRID_Z/2);
				glVertex3f(x-GR_GRID_X/2,y1-GR_GRID_Y/2,z-GR_GRID_Z/2);
			}
			for (int x = GR_GRID_X - 1; x >= 0; x--) {
				int y2 = 0;
				for (int y = GR_GRID_Y/2; y >= 0; y--) {
					if (grid[x][y][z].cell.celltype == GRID_NOGO) {
						y2 = y;
						break;
					}
				}
				if (x != 0)
					glVertex3f(x-GR_GRID_X/2,y2-GR_GRID_Y/2,z-GR_GRID_Z/2);
				glVertex3f(x-GR_GRID_X/2,y2-GR_GRID_Y/2,z-GR_GRID_Z/2);
			}
		}
#endif

#ifdef DISPGRIDY
		// now iterate through x with z > GR_GRID_Z/2 and z < GR_GRID_Z/2 and find min z
		for (int y = 0; y < GR_GRID_Y; y++) {
			for (int x = 0; x < GR_GRID_X; x++) {
				int z1 = GR_GRID_Z - 1;
				for (int z = GR_GRID_Z/2; z < GR_GRID_Z; z++) {
					if (grid[x][y][z].elementtype == ELEMENT_OPEN) {
						z1 = z;
						break;
					}
				}
				if (x != 0)
					glVertex3f(x-GR_GRID_X/2,y-GR_GRID_Y/2,z1-GR_GRID_Z/2);
				glVertex3f(x-GR_GRID_X/2,y-GR_GRID_Y/2,z1-GR_GRID_Z/2);
			}
			for (int x = GR_GRID_X - 1; x >= 0; x--) {
				int z2 = 0;
				for (int z = GR_GRID_Z/2; z >= 0; z--) {
					if (grid[x][y][z].elementtype == ELEMENT_OPEN) {
						z2 = z;
						break;
					}
				}
				if (x != 0)
					glVertex3f(x-GR_GRID_X/2,y-GR_GRID_Y/2,z2-GR_GRID_Z/2);
				glVertex3f(x-GR_GRID_X/2,y-GR_GRID_Y/2,z2-GR_GRID_Z/2);
			}
		}
#endif

		glEnd();
	}
/*
#ifndef SHOWALLCELLS
#ifdef DRAWPOINTS
	glBegin(GL_POINTS);
	for (int x = 0; x < GR_GRID_X; x++) {
		for (int y = 0; y < GR_GRID_Y; y++) {
			for (int z = 0; z < GR_GRID_Z; z++) {
				if (!grid[x][y][z].cell.displaybig && grid[x][y][z].cell.celltype != CELL_NONE
						&& grid[x][y][z].cell.celltype != GRID_NOGO
						&& grid[x][y][z].cell.celltype != GRID_FRC) {
					if (grid[x][y][z].cell.celltype == CELL_K)
						glColor3f(0.5f, 0, 0);
					else if (grid[x][y][z].cell.celltype == CELL_H)
						glColor3f(0, 0, 0.5f);
			//		else if (grid[x][y][z].cell.celltype == GRID_FRC)
			//			glColor3f(1.0f, 1.0f, 1.0f);
				//	else if (grid[x][y][z].cell.celltype == CELL_D)
				//		glColor3f(0, 1.0f, 0);
				//	else if (grid[x][y][z].cell.celltype == GRID_NOGO)
				//		glColor3f(1.0, 1.0, 1.0);
	//				glColor4f(0.0f,1.0f,0.0f,0.5f);
					// these are infinitely small so allow +-0.5 unit
					float xrand = 0, yrand = 0, zrand = 0;
			//		if (grid[x][y][z].cell.celltype != GRID_FRC) {
						xrand = rnd.getfloat() - 0.5f;
						yrand = rnd.getfloat() - 0.5f;
						zrand = rnd.getfloat() - 0.5f;
			//		}
					glVertex3f(x - GR_GRID_X/2 + xrand, y - GR_GRID_Y/2 + yrand, z - GR_GRID_Z/2 + zrand);
				}
			}
		}
	}
	glEnd();
#endif
#endif
*/
	glEnable(GL_LIGHTING);
	for (int x = 0; x < GR_GRID_X; x++) {
		for (int y = 0; y < GR_GRID_Y; y++) {
			for (int z = 0; z < GR_GRID_Z; z++) {
#ifndef SHOWALLCELLS
				if (grid[x][y][z].cell.displaybig && grid[x][y][z].cell.celltype != CELL_NONE && grid[x][y][z].cell.celltype != GRID_NOGO) {
#else
				if (grid[x][y][z].cell.celltype != LCELL_NONE && grid[x][y][z].cell.celltype != LCELL_NOGO) {
#endif
					glPushMatrix();
					// need to finalize colors
					if (grid[x][y][z].cell.celltype == CELL_TREG) {
						if (grid[x][y][z].cell.state == TREGCELL_RESTING)
							glColor3f(0.2f, 0.4f, 1.0f); // light blue
						else if (grid[x][y][z].cell.state == TREGCELL_ACTIVE)
							glColor3f(0.2f, 1.0f, 0.6f); // aqua
						else if (grid[x][y][z].cell.state == TREGCELL_REG)
							glColor3f(0.2f, 1.0f, 0.6f); // aqua
					} else if (grid[x][y][z].cell.celltype == CELL_CD4) {
						if (grid[x][y][z].cell.state == CD4CELL_RESTING)
							glColor3f(0.2f, 0.4f, 1.0f); // light blue
						else if (grid[x][y][z].cell.state == CD4CELL_ACTIVE)
							glColor3f(0.2f, 1.0f, 0.6f); // aqua
					} else if (grid[x][y][z].cell.celltype == CELL_CD8) {
						if (grid[x][y][z].cell.state == CD8CELL_RESTING)
							glColor3f(1.0f, 0.2f, 0.2f); // light red
						else if (grid[x][y][z].cell.state == CD8CELL_ACTIVE)
							glColor3f(1.0f, 0.2f, 1.0f); // purple
					} else if (grid[x][y][z].cell.celltype == CELL_D) {
						if (grid[x][y][z].cell.state == DCELL_IMMATURE)
							glColor3f(0.0f, 1.0f, 0.0f); // green
						else if (grid[x][y][z].cell.state == DCELL_MATURE)
							glColor3f(0.4f, 1.0f, 0.4f); // light green
						else if (grid[x][y][z].cell.state == DCELL_LICENSED)
							glColor3f(1.0f, 1.0f, 1.0f); // white
					} else if (grid[x][y][z].cell.celltype == CELL_M) {
						if (grid[x][y][z].cell.state == MCELL_RESTING)
							glColor3f(0.2f, 0.4f, 1.0f); // light blue
							//glColor3f(0.0f, 1.0f, 0.0f); // green
						else if (grid[x][y][z].cell.state == MCELL_INFECTED
						  || grid[x][y][z].cell.state == MCELL_CINFECTED)
							glColor3f(1.0f, 0.2f, 0.2f); // light red
							//glColor3f(0.4f, 1.0f, 0.4f); // light green
						else if (grid[x][y][z].cell.state == MCELL_ACTIVE)
							glColor3f(1.0f, 1.0f, 1.0f); // white
					}
					float specReflection[] = { 1.0f, 1.0f, 1.0f, 1.0f };
					glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specReflection);
					glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, 200);

					// since the spheres are 0.35 radius = 0.7 diameter, they can budge by +- 0.15 in each axis
					// generate some offset values
					float xrand, yrand, zrand;
					#ifndef SHOWALLCELLS
						xrand = yrand = zrand = 0;
					#else
						if (domove == 0 || domove == 1) {
							xrand = yrand = zrand = 0;
						} else {
							xrand = getrand() / 2.5f - 0.2f;
							yrand = getrand() / 2.5f - 0.2f;
							zrand = getrand() / 2.5f - 0.2f;
						}
					#endif

					glTranslatef(x - GR_GRID_X/2 + xrand, y - GR_GRID_Y/2 + yrand, z - GR_GRID_Z/2 + zrand);
					if (grid[x][y][z].cell.celltype == CELL_M || grid[x][y][z].cell.celltype == CELL_D)
						glTranslatef(0.5, 0.5, 0.5);
					glPushMatrix();
					if (grid[x][y][z].cell.celltype == CELL_TREG || grid[x][y][z].cell.celltype == CELL_CD4 || grid[x][y][z].cell.celltype == CELL_CD8 || grid[x][y][z].cell.celltype == CELL_D || grid[x][y][z].cell.celltype == CELL_M)
						glRotatef(directions3[grid[x][y][z].cell.lastdir][0],directions3[grid[x][y][z].cell.lastdir][1],
							directions3[grid[x][y][z].cell.lastdir][2],directions3[grid[x][y][z].cell.lastdir][3]);
					//glRotatef(theta, 0, 1, 0); // for spinning
					if (grid[x][y][z].cell.celltype == CELL_D || grid[x][y][z].cell.celltype == CELL_M) {
					//	glScalef(1.0f, 2.0f, 1.0f);
						glCallList(mysphereIDbig);
//					} else if (grid[x][y][z].cell.celltype == OUTPUT_EFFERENT || grid[x][y][z].cell.celltype == INPUT_AFFERENT || grid[x][y][z].cell.celltype == INPUT_HEV) {
//						glCallList(mycylID);
					} else {
					//	glScalef(1.0f, 2.0f, 1.0f);
						glCallList(mysphereID);
					}
					glPopMatrix();

					glPopMatrix();
				}
			}
		}
	}

	glPopMatrix();

//	glPushMatrix();
	//glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_LIGHTING);
	glLoadIdentity();
	glTranslatef(0.0f, 0.0f, -1.0f);
	// now draw text
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	// Position The Text On The Screen
	glRasterPos2f(0.46, 0.38);

// Disable Windows specifc code, while trying to get the code to run under Linux.
#ifdef CLOCK
 	glPrint("%0.1fh", (float)timeindex/600.0f);	// Print GL Text To The Screen
#endif

	glEnable(GL_LIGHTING);
}

void GRGrid::Recruitment() {
}

void GRGrid::Movement() {
	vector<coordc> neighbor;
	vector<float> neighbor_sum;
	for (int x = 0; x < GR_GRID_X; x++) {
		for (int y = 0; y < GR_GRID_Y; y++) {
			for (int z = 0; z < GR_GRID_Z; z++) {
				if (grid[x][y][z].cell.celltype == CELL_CD4 || 
					grid[x][y][z].cell.celltype == CELL_CD8 || 
					grid[x][y][z].cell.celltype == CELL_TREG || 
					grid[x][y][z].cell.celltype == CELL_M) {
				//	float tnf = bilinear_3d(coordc(x,y,z), CHEM_TNF);
					float ccl2 = bilinear_3d(coordc(x,y,z), CHEM_CCL2);
					float ccl5 = bilinear_3d(coordc(x,y,z), CHEM_CCL5);
					float cxcl9 = bilinear_3d(coordc(x,y,z), CHEM_CXCL9);
				//	float extmtb = bilinear_3d(coordc(x,y,z), CHEM_EXTMTB);

					// compute neighbors
					neighbor.clear();
					neighbor_sum.clear();
					for (int p = max(0, x - 1); p <= min(GR_GRID_X, x + 1); p++) {
						for (int q = max(0, y - 1); q <= min(GR_GRID_Y, y + 1); q++) {
							for (int r = max(0, z - 1); r <= min(GR_GRID_Z, z + 1); r++) {
								float csum = 1.0f;
								if (grid[x][y][z].cell.celltype == CELL_CD4) {
									if (ccl2 > MIN_CCL2_CD4_TCELL && ccl2 < SAT_CCL2_CD4_TCELL)
										csum += bilinear_3d(coordc(p,q,r), CHEM_CCL2);
									if (ccl5 > MIN_CCL5_CD4_TCELL && ccl5 < SAT_CCL5_CD4_TCELL)
										csum += bilinear_3d(coordc(p,q,r), CHEM_CCL5);
									if (cxcl9 > MIN_CXCL9_CD4_TCELL && cxcl9 < SAT_CXCL9_CD4_TCELL)
										csum += bilinear_3d(coordc(p,q,r), CHEM_CXCL9);
								} else if (grid[x][y][z].cell.celltype == CELL_CD8) {
									if (ccl5 > MIN_CCL5_CD8_TCELL && ccl5 < SAT_CCL5_CD8_TCELL)
										csum += bilinear_3d(coordc(p,q,r), CHEM_CCL5);
									if (cxcl9 > MIN_CXCL9_CD8_TCELL && cxcl9 < SAT_CXCL9_CD8_TCELL)
										csum += bilinear_3d(coordc(p,q,r), CHEM_CXCL9);
								} else if (grid[x][y][z].cell.celltype == CELL_TREG) {
									if (ccl5 > MIN_CCL5_REG_TCELL && ccl5 < SAT_CCL5_REG_TCELL)
										csum += bilinear_3d(coordc(p,q,r), CHEM_CCL5);
								} else if (grid[x][y][z].cell.celltype == CELL_M) { // need to average 8 for this
									if (ccl2 > MIN_CCL2_MAC && ccl2 < SAT_CCL2_MAC)
										csum += bilinear_3d(coordc(p,q,r), CHEM_CCL2);
									if (ccl5 > MIN_CCL5_MAC && ccl5 < SAT_CCL5_MAC)
										csum += bilinear_3d(coordc(p,q,r), CHEM_CCL5);
								}
								neighbor.push_back(coordc(p,q,r));
								neighbor_sum.push_back(csum);
							}
						}
					}
					
					// now normalize
					float sum = 0.0f, themax = 0.0f;
					int themaxidx = 0;
					for (int p = 0; p < neighbor.size(); p++)
						if (neighbor_sum[p] > themax) {
							themax = neighbor_sum[p];
							themaxidx = p;
						}
					neighbor_sum[themaxidx] *= 1.5f; // strong tendency to follow main gradient
					for (int p = 0; p < neighbor.size(); p++)
						sum += neighbor_sum[p];
					for (int p = 0; p < neighbor.size(); p++)
						neighbor_sum[p] /= sum;

					float floatchoice = rnd.getfloat(), tempsum = 0.0f;
					int dir = 0;
					for (dir = 0; dir < neighbor.size(); dir++) {
						tempsum += neighbor_sum[dir];
						if (floatchoice < tempsum)
							break;
						// else keep going until we reach the random number
					}
					coordc newpos = neighbor[dir];
					if (grid[newpos.x][newpos.y][newpos.z].cell.celltype == CELL_NONE &&
						1 /* caseation and maybe other constraints?!? */) { // move the cell
						grid[newpos.x][newpos.y][newpos.z].cell = grid[x][y][z].cell;
						grid[x][y][z].cell.celltype = CELL_NONE;
					} // else do nothing
				}
			}
		}
	}

/*	// MOVE CELLS
	for (int x = 0; x < GR_GRID_X; x++) {
		for (int y = 0; y < GR_GRID_Y; y++) {
			for (int z = 0; z < GR_GRID_Z; z++) {
				// generate direction and validate
				if (grid[x][y][z].cell.movedyet == false && grid[x][y][z].cell.celltype != CELL_NONE) {
					if ((grid[x][y][z].cell.celltype == CELL_D || grid[x][y][z].cell.celltype == CELL_M) && timeindex % 40 != 0) // don't do DCs every cycle
						continue;
					int numdirs = getnumdirs(grid[x][y][z].cell.lastdir);
					int lastdir = grid[x][y][z].cell.lastdir;
					int newdir = rand() % numdirs;
					newdir = nearbydirs[grid[x][y][z].cell.lastdir][newdir];
					coord current;
					current.x = x;
					current.y = y;
					current.z = z;
					coord newcoord = applydir(current, newdir);
					//coord newcoord = applydir(current, grid[x][y][z].cell.lastdir);
					grid[x][y][z].cell.lastdir = newdir;
					grid[x][y][z].cell.movedyet = true;
					if (grid[newcoord.x][newcoord.y][newcoord.z].cell.celltype == CELL_NONE) {
						grid[newcoord.x][newcoord.y][newcoord.z].cell = grid[x][y][z].cell;
						grid[x][y][z].cell.celltype = CELL_NONE; // remove old
						// did move, so increment
//						if (newdir == lastdir) {
//							grid[newcoord.x][newcoord.y][newcoord.z].cell.mfp += MFP_UNIT;
//						} else { // save last mfp
//							mfpsum += grid[x][y][z].cell.mfp;
//							nummfp++;
//							grid[newcoord.x][newcoord.y][newcoord.z].cell.mfp = MFP_UNIT;
//						}
					}
					//else if (grid[newcoord.x][newcoord.y][newcoord.z].cell.celltype == OUTPUT_EFFERENT) { // remove
					//	mfpsum += grid[x][y][z].cell.mfp; // save last mfp
					//	nummfp++;

					//	totalgridtime += timeindex - grid[x][y][z].cell.enter_time; // to figure out how long cells stay on the grid
					//	totalcellslost++;
					//	totalcells[grid[x][y][z].cell.celltype]--;
					//	if (totalcellslost >= 30000) {
					//		totalgridtime /= 2;
					//		totalcellslost /= 2;
					//	}
					//	grid[x][y][z].cell.celltype = CELL_NONE;
					//} // if no movement maybe clear mfp?
				}
			}
		}
	}*/
}

void GRGrid::ClearMoved() {
	for (int x = 0; x < GR_GRID_X; x++) {
		for (int y = 0; y < GR_GRID_Y; y++) {
			for (int z = 0; z < GR_GRID_Z; z++) {
				if (grid[x][y][z].cell.celltype == CELL_TREG || grid[x][y][z].cell.celltype == CELL_CD4 || grid[x][y][z].cell.celltype == CELL_CD8)
					grid[x][y][z].cell.movedyet = false;
				if ((grid[x][y][z].cell.celltype == CELL_D || grid[x][y][z].cell.celltype == CELL_M))// && grid[x][y][z].cell.lockdown == 0)
					grid[x][y][z].cell.movedyet = false;
			}
		}
	}
}

void GRGrid::States() {
	for (int x = 0; x < GR_GRID_X; x++) {
		for (int y = 0; y < GR_GRID_Y; y++) {
			for (int z = 0; z < GR_GRID_Z; z++) {
				grid[x][y][z].cell.next_state = grid[x][y][z].cell.state; // just in case
			}
		}
	}
	for (int x = 0; x < GR_GRID_X; x++) {
		for (int y = 0; y < GR_GRID_Y; y++) {
			for (int z = 0; z < GR_GRID_Z; z++) {
				grid[x][y][z].cell.state = grid[x][y][z].cell.next_state; // implement
			}
		}
	}
}

void GRGrid::ClearDead() {
	for (int x = 0; x < GR_GRID_X; x++) {
		for (int y = 0; y < GR_GRID_Y; y++) {
			for (int z = 0; z < GR_GRID_Z; z++) {
				if ((grid[x][y][z].cell.celltype == CELL_TREG || 
					grid[x][y][z].cell.celltype == CELL_CD4 || 
					grid[x][y][z].cell.celltype == CELL_CD8 || 
					grid[x][y][z].cell.celltype == CELL_D || 
					grid[x][y][z].cell.celltype == CELL_M) && 
					grid[x][y][z].cell.state == CELL_DEAD) {
						// remove and possibly save some statistics
						grid[x][y][z].cell.celltype = CELL_NONE;
				}
			}
		}
	}
}


bool GRGrid::InMoore(coordc currloc, CellTypes tofind, coord * loc) {
	// iterate through +- 1 cell in each direction, return when we find something
	int startx = currloc.x - 1;
	int endx = currloc.x + 1;
	int starty = currloc.y - 1;
	int endy = currloc.y + 1;
	int startz = currloc.z - 1;
	int endz = currloc.z + 1;

	if (startx < 0)
		startx = 0;
	if (starty < 0)
		starty = 0;
	if (startz < 0)
		startz = 0;
	if (endx >= GR_GRID_X)
		endx = GR_GRID_X - 1;
	if (endy >= GR_GRID_Y)
		endy = GR_GRID_Y - 1;
	if (endz >= GR_GRID_Z)
		endz = GR_GRID_Z - 1;

	for (int x = startx; x <= endx; x++) {
		for (int y = starty; y <= endy; y++) {
			for (int z = startz; z <= endz; z++) {
				if (grid[x][y][z].cell.celltype == tofind &&
					!(x == currloc.x && y == currloc.y && z == currloc.z)) {
					loc->x = x;
					loc->y = y;
					loc->z = z;
					return true;
				}
			}
		}
	}
	return false;
}

bool GRGrid::InMooreState(coordc currloc, CellTypes tofind, CellStates tomatch, coord * loc) {
	// iterate through +- 1 cell in each direction, return when we find something
	int startx = currloc.x - 1;
	int endx = currloc.x + 1;
	int starty = currloc.y - 1;
	int endy = currloc.y + 1;
	int startz = currloc.z - 1;
	int endz = currloc.z + 1;

	if (startx < 0)
		startx = 0;
	if (starty < 0)
		starty = 0;
	if (startz < 0)
		startz = 0;
	if (endx >= GR_GRID_X)
		endx = GR_GRID_X - 1;
	if (endy >= GR_GRID_Y)
		endy = GR_GRID_Y - 1;
	if (endz >= GR_GRID_Z)
		endz = GR_GRID_Z - 1;

	for (int x = startx; x <= endx; x++) {
		for (int y = starty; y <= endy; y++) {
			for (int z = startz; z <= endz; z++) {
				if (grid[x][y][z].cell.celltype == tofind && grid[x][y][z].cell.state == tomatch &&
					!(x == currloc.x && y == currloc.y && z == currloc.z)) {
					loc->x = x;
					loc->y = y;
					loc->z = z;
					return true;
				}
			}
		}
	}
	return false;
}

bool GRGrid::InMooreStateRadius(coordc currloc, CellTypes tofind, CellStates tomatch, float radius, coord * loc) {
	// iterate through +- 1 cell in each direction, return when we find something
	int radmax = ceil(radius);
	int startx = currloc.x - radmax;
	int endx = currloc.x + radmax;
	int starty = currloc.y - radmax;
	int endy = currloc.y + radmax;
	int startz = currloc.z - radmax;
	int endz = currloc.z + radmax;

	if (startx < 0)
		startx = 0;
	if (starty < 0)
		starty = 0;
	if (startz < 0)
		startz = 0;
	if (endx >= GR_GRID_X)
		endx = GR_GRID_X - 1;
	if (endy >= GR_GRID_Y)
		endy = GR_GRID_Y - 1;
	if (endz >= GR_GRID_Z)
		endz = GR_GRID_Z - 1;

	for (int x = startx; x <= endx; x++) {
		for (int y = starty; y <= endy; y++) {
			for (int z = startz; z <= endz; z++) {
				float dist = sqrt(pow((float)(x-currloc.x),2) + pow((float)(y-currloc.y),2) + pow((float)(z-currloc.z),2));
				if (dist <= radius && grid[x][y][z].cell.celltype == tofind && grid[x][y][z].cell.state == tomatch &&
					!(x == currloc.x && y == currloc.y && z == currloc.z)) {
					loc->x = x;
					loc->y = y;
					loc->z = z;
					return true;
				}
			}
		}
	}
	return false;
}

coordc GRGrid::GenEmptyCoord(coordc currloc, int delta) {
	int startx = currloc.x - delta;
	int endx = currloc.x + delta;
	int starty = currloc.y - delta;
	int endy = currloc.y + delta;
	int startz = currloc.z - delta;
	int endz = currloc.z + delta;
	coordc out(0,0,0);

	if (delta == -1) {
		while (1) {
			out.x = rnd.get32() % GR_GRID_X;
			out.y = rnd.get32() % GR_GRID_Y;
			out.z = rnd.get32() % GR_GRID_Z;
			if (grid[out.x][out.y][out.z].elementtype != ELEMENT_OPEN && grid[out.x][out.y][out.z].cell.celltype == CELL_NONE)
				return out;
		}
	}

	if (startx < 0)
		startx = 0;
	if (starty < 0)
		starty = 0;
	if (startz < 0)
		startz = 0;
	if (endx >= GR_GRID_X)
		endx = GR_GRID_X - 1;
	if (endy >= GR_GRID_Y)
		endy = GR_GRID_Y - 1;
	if (endz >= GR_GRID_Z)
		endz = GR_GRID_Z - 1;

	while (1) {
		out.x = rnd.get_inrange(startx, endx);
		out.y = rnd.get_inrange(starty, endy);
		out.z = rnd.get_inrange(startz, endz);

		if (grid[out.x][out.y][out.z].elementtype != ELEMENT_OPEN && grid[out.x][out.y][out.z].cell.celltype == CELL_NONE)
			return out;
	}
}

int GRGrid::ClearSphere(coordc currloc, float radius) {
	int radmax = ceil(radius);
	int startx = currloc.x - radmax;
	int endx = currloc.x + radmax;
	int starty = currloc.y - radmax;
	int endy = currloc.y + radmax;
	int startz = currloc.z - radmax;
	int endz = currloc.z + radmax;
	int total = 0;

	if (startx < 0)
		startx = 0;
	if (starty < 0)
		starty = 0;
	if (startz < 0)
		startz = 0;
	if (endx >= GR_GRID_X)
		endx = GR_GRID_X - 1;
	if (endy >= GR_GRID_Y)
		endy = GR_GRID_Y - 1;
	if (endz >= GR_GRID_Z)
		endz = GR_GRID_Z - 1;

	for (int x = startx; x <= endx; x++) {
		for (int y = starty; y <= endy; y++) {
			for (int z = startz; z <= endz; z++) {
				float dist = sqrt(pow((float)(x-currloc.x),2) + pow((float)(y-currloc.y),2) + pow((float)(z-currloc.z),2));
				if (dist <= radius) {
					grid[x][y][z].elementtype = ELEMENT_OPEN;
					total++;
				}
			}
		}
	}

	return total;
}

