
#include "stdafx.h"
using namespace std;

#define WRITE_NUMCELLS	0
#define WRITE_PNG		0
#define WRITE_VIDEO		0
#define DISPLAY_SDL		1

#define OVERLAY_TYPE	SDL_UYVY_OVERLAY
#define OVERLAY_FMT		PIX_FMT_UYVY422
#define VIDEO_FMT		PIX_FMT_YUV420P

#define PX_PER_CELL		4
#define IMAGE_WIDTH		(GRID_X * 2 * PX_PER_CELL)
#define IMAGE_HEIGHT	(GRID_Y * 2 * PX_PER_CELL)

png_byte *row_pointers[IMAGE_HEIGHT];
png_color image[IMAGE_HEIGHT][IMAGE_WIDTH];

static CompareStruct comparegrid[GRID_X][GRID_Y][GRID_Z][SLOTS]; // 1 byte * 20000 = 20KB
static GridCoord comparehevs[NUM_HEV]; // 3 bytes * 40 = 120B
static GridCoord compareafferents[NUM_AFFERENT]; // 3 bytes * 100 = 300B
static GridCoord compareefferents[NUM_EFFERENT]; // 3 bytes * 160 = 480B
OutStruct temp;
u32 framenumber = 0, lasthev = 0, lastafferent = 0, lastefferent = 0;
png_color bgcolor = {0, 0, 0}, whitecolor = {255, 255, 255}, alertcolor = {255, 105, 180};
s32 cellcounts[NUMCELLTYPES] = {0};

void perspectiveGL(GLdouble fovY, GLdouble aspect, GLdouble zNear, GLdouble zFar);
GLfloat clamp(GLfloat value, GLfloat min, GLfloat max);
#ifdef WIN32
FILE * globalfd;
size_t rbt_fwrite(int fd, const void *buf, size_t count) {
	return fwrite(buf, 1, count, globalfd); }
int rbt_fflush(FILE * stream) {	return fflush(stream); }
#endif

#ifdef WIN32
int _tmain(int argc, _TCHAR* argv[]) {
#else
int main (int argc, char *argv[]) {
#endif
	assert(sizeof(OutStruct) == 4);

	ifstream infile;
	if (argc < 2) {
		cout << "Error, no file specified, using default." << endl;
		infile.open("LN_example.dat", ios::in | ios::binary);
	} else
		infile.open(argv[1], ios::in | ios::binary);

	if (!infile.is_open()) {
		cout << "Error, cannot open file: " << argv[1] << endl;
		return 1;
	}

#ifdef WIN32
	system("mkdir images");
#else
	system("mkdir images 2>/dev/null");
#endif
	ofstream numcellsfile;

	if (WRITE_NUMCELLS) {
		//enum CellTypes { CELL_NONE = 0, CELL_ALL = 0, CELL_H = 1, CELL_K = 2, CELL_D = 3,
		//CELL_TG = 4, CELL_TC = 5, CELL_TR = 6, CELL_M = 7 }; // 8
		numcellsfile.open("numcells.csv");//, ios::out | out::trunc); // weird error with ios::out
		numcellsfile << "Time (hrs),# None Cells,# Helper T Cells,# Killer T Cells,# Dendritic Cells," <<
			"# TG Cells,# TC Cells,# TR Cells,# Macrophages,# Afferent Nodes,# Efferent Nodes" << endl;
	}

    avcodec_init();
    avcodec_register_all();

	SDL_Surface *screen;
	SDL_Rect rect;
	SDL_Event event;
	SDL_Overlay *bmp;
	AVFrame * pFrameRGB;
	struct SwsContext *img_convert_ctx = NULL, *video_convert_ctx = NULL;
	if (DISPLAY_SDL || WRITE_VIDEO) {
		if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER)) { //  | SDL_INIT_AUDIO
		  fprintf(stderr, "Could not initialize SDL - %s\n", SDL_GetError());
		  exit(1);
		}

		if (DISPLAY_SDL) {
			screen = SDL_SetVideoMode(IMAGE_WIDTH, IMAGE_HEIGHT, 0, 0);
			if (!screen) {
				fprintf(stderr, "SDL: could not set video mode - exiting\n");
			exit(1);
			}
		}

		pFrameRGB = avcodec_alloc_frame();
		avpicture_fill((AVPicture *)pFrameRGB, (uint8_t *)&image, PIX_FMT_RGB24, IMAGE_WIDTH, IMAGE_HEIGHT);
		assert(sizeof(image) == avpicture_get_size(PIX_FMT_RGB24, IMAGE_WIDTH, IMAGE_HEIGHT));

		if (DISPLAY_SDL) {
			bmp = SDL_CreateYUVOverlay(IMAGE_WIDTH, IMAGE_HEIGHT, OVERLAY_TYPE, screen);

			img_convert_ctx = sws_getContext(
				IMAGE_WIDTH, IMAGE_HEIGHT, PIX_FMT_RGB24,
				IMAGE_WIDTH, IMAGE_HEIGHT, OVERLAY_FMT,
				SWS_BICUBIC, NULL, NULL, NULL);
			if (img_convert_ctx == NULL) {
				fprintf(stderr, "Cannot initialize the conversion context!\n");
				exit(1);
			}
		}
		if (WRITE_VIDEO) {
			video_convert_ctx = sws_getContext(
				IMAGE_WIDTH, IMAGE_HEIGHT, PIX_FMT_RGB24,
				IMAGE_WIDTH, IMAGE_HEIGHT, VIDEO_FMT,
				SWS_BICUBIC, NULL, NULL, NULL);
			if (video_convert_ctx == NULL) {
				fprintf(stderr, "Cannot initialize the conversion context!\n");
				exit(1);
			}
		}
	}


    AVCodec *codec;
    AVCodecContext *c= NULL;
    int i = 0, out_size, size, outbuf_size;
    FILE *f;
    AVFrame *picture;
    uint8_t *outbuf, *picture_buf;
	if (WRITE_VIDEO) {
		codec = avcodec_find_encoder(CODEC_ID_MPEG2VIDEO); //CODEC_ID_FLV1 CODEC_ID_MPEG1VIDEO CODEC_ID_MPEG4
		if (!codec) {
			fprintf(stderr, "codec not found\n");
			exit(1);
		}

		c= avcodec_alloc_context();
		picture= avcodec_alloc_frame();

		/* put sample parameters */
		c->bit_rate = 2000000;
		/* resolution must be a multiple of two */
		c->width = IMAGE_WIDTH;
		c->height = IMAGE_HEIGHT;
		/* frames per second */
		//c->time_base= (AVRational){1,25};
		c->time_base.den = 25;// = (AVRational){1,25};
		c->time_base.num = 1;
		c->gop_size = 10; /* emit one intra frame every ten frames */
		//c->max_b_frames=1; // MPEG1
        c->max_b_frames = 2; // MPEG2
		c->pix_fmt = VIDEO_FMT;
		c->qmax = 15;
		c->qmin = 1;

		/* open it */
		if (avcodec_open(c, codec) < 0) {
			fprintf(stderr, "could not open codec\n");
			exit(1);
		}

		char filename[100] = "video.mpg";
		f = fopen(filename, "wb");
		if (!f) {
			fprintf(stderr, "could not open %s\n", filename);
			exit(1);
		}

		/* alloc image and output buffer */
		outbuf_size = 10000000; // 10 MB
		outbuf = (uint8_t *)malloc(outbuf_size);
		size = c->width * c->height;
		picture_buf = (uint8_t *)malloc((size * 3) / 2); /* size for YUV 420 */
		//picture_buf = (uint8_t *)malloc((size * 3)); /* overallocate size for YUV 420 */

		picture->data[0] = picture_buf;
		picture->data[1] = picture->data[0] + size;
		picture->data[2] = picture->data[1] + size / 4;
		//picture->data[2] = picture->data[1] + size;
		picture->linesize[0] = c->width;
		//picture->linesize[1] = c->width;
		//picture->linesize[2] = c->width;
		picture->linesize[1] = c->width / 2;
		picture->linesize[2] = c->width / 2;
	}

    for (int r = 0; r < IMAGE_HEIGHT; r++) {
    	row_pointers[r] = (png_byte *)image[r];
    	for (int c = 0; c < IMAGE_WIDTH; c++) {
    	//	image[r][c].red = c;
    	//	image[r][c].green = 255 - r;
    	//	image[r][c].blue = r;
    		image[r][c] = bgcolor;
    	}
    }

	// clear struct
	for (int x = 0; x < GRID_X; x++)
		for (int y = 0; y < GRID_Y; y++)
			for (int z = 0; z < GRID_Z; z++)
				for (int s = 0; s < 4; s++) {
					comparegrid[x][y][z][s].type = CELL_NONE;
					comparegrid[x][y][z][s].mode = 0;
				}

	// now start reading 4-byte blocks and applying the changes until we have a complete frame
	while (!infile.eof()) {
		do {
			png_color cellcolor, blendcolor;
			infile.read((char *)&temp, sizeof(temp));

			//cout << "RECORD: x" << temp.x << " y" << temp.y << " z" << temp.z << " s" << temp.s
			//	<< " type" << temp.type << " state" << temp.state << " mode" << temp.mode << endl;

			if (temp.mode == 0 && temp.x < GRID_X && temp.y < GRID_Y && temp.z < GRID_Z) { // apply to comparegrid
				cellcounts[comparegrid[temp.x][temp.y][temp.z][temp.s].type]--;
				cellcounts[temp.type]++;

				comparegrid[temp.x][temp.y][temp.z][temp.s].mode = temp.mode;
				comparegrid[temp.x][temp.y][temp.z][temp.s].state = (CellStates)temp.state;
				comparegrid[temp.x][temp.y][temp.z][temp.s].type = (CellTypes)temp.type;

				u32 height_idx_start = (temp.y * 2 + temp.s/2) * PX_PER_CELL;
				u32 width_idx_start = (temp.x * 2 + temp.s % 2) * PX_PER_CELL;

				if (comparegrid[temp.x][temp.y][temp.z][temp.s].type == CELL_H) {
					cellcolor.red = 0;
					cellcolor.green = 180;
					cellcolor.blue = 255;
				} else if (comparegrid[temp.x][temp.y][temp.z][temp.s].type == CELL_K) {
					cellcolor.red = 255;
					cellcolor.green = 0;
					cellcolor.blue = 0;
				} else if (comparegrid[temp.x][temp.y][temp.z][temp.s].type == CELL_D) {
					cellcolor.red = 0;
					cellcolor.green = 255;
					cellcolor.blue = 0;
				} else if (comparegrid[temp.x][temp.y][temp.z][temp.s].type == CELL_NONE) {
					cellcolor = bgcolor;
				} else {
					// wtf?
					cellcolor = alertcolor;
				}

				for (int h = 0; h < PX_PER_CELL; h++)
					for (int w = 0; w < PX_PER_CELL; w++) {
						//blendcolor = whitecolor;
						blendcolor = bgcolor;
						if (h != 0)
							if (w != 0)
								blendcolor = cellcolor;
						if ((h == 1 || h == PX_PER_CELL - 1) && (w == 1 || w == PX_PER_CELL - 1)) {
							blendcolor.red = (5 * (int)cellcolor.red) / 10 + (5 * (int)bgcolor.red) / 10;
							blendcolor.green = (5 * (int)cellcolor.green) / 10 + (5 * (int)bgcolor.green) / 10;
							blendcolor.blue = (5 * (int)cellcolor.blue) / 10 + (5 * (int)bgcolor.blue) / 10;
							//blendcolor = bgcolor;
						}

						/*if (h == 1 || h == 2 || h == 3 || h == 4 || h == 5) {
							if (w == 1 || w == 2 || w == 3 || w == 4 || w == 5) {
								blendcolor = cellcolor;
							} else if (h == 1 || h == 4) { // 0.3
								blendcolor.red = (3 * (int)cellcolor.red) / 10 + (7 * (int)bgcolor.red) / 10;
								blendcolor.green = (3 * (int)cellcolor.green) / 10 + (7 * (int)bgcolor.green) / 10;
								blendcolor.blue = (3 * (int)cellcolor.blue) / 10 + (7 * (int)bgcolor.blue) / 10;
							} else if (h == 2 || h == 3) { // 0.8
								blendcolor.red = (8 * (int)cellcolor.red) / 10 + (2 * (int)bgcolor.red) / 10;
								blendcolor.green = (8 * (int)cellcolor.green) / 10 + (2 * (int)bgcolor.green) / 10;
								blendcolor.blue = (8 * (int)cellcolor.blue) / 10 + (2 * (int)bgcolor.blue) / 10;
							}
						} else {
							if (w == 0 || w == 5) {
								blendcolor = bgcolor;
							} else if (w == 1 || w == 4) { // 0.3
								blendcolor.red = (3 * (int)cellcolor.red) / 10 + (7 * (int)bgcolor.red) / 10;
								blendcolor.green = (3 * (int)cellcolor.green) / 10 + (7 * (int)bgcolor.green) / 10;
								blendcolor.blue = (3 * (int)cellcolor.blue) / 10 + (7 * (int)bgcolor.blue) / 10;
							} else if (w == 2 || w == 3) { // 0.8
								blendcolor.red = (8 * (int)cellcolor.red) / 10 + (2 * (int)bgcolor.red) / 10;
								blendcolor.green = (8 * (int)cellcolor.green) / 10 + (2 * (int)bgcolor.green) / 10;
								blendcolor.blue = (8 * (int)cellcolor.blue) / 10 + (2 * (int)bgcolor.blue) / 10;
							}
						}*/

						// FOR PX_PER_CELL = 5
						/*if (h == 1 || h == 2 || h == 3) {
							if (w == 1 || w == 2 || w == 3) {
								blendcolor = cellcolor;
							} else if (w == 0 || w == 4) {
								if (h == 1 || h == 3) {
									blendcolor.red = (5 * (int)cellcolor.red) / 10 + (5 * (int)bgcolor.red) / 10;
									blendcolor.green = (5 * (int)cellcolor.green) / 10 + (5 * (int)bgcolor.green) / 10;
									blendcolor.blue = (5 * (int)cellcolor.blue) / 10 + (5 * (int)bgcolor.blue) / 10;
								} else if (h == 2) { // 2
									blendcolor.red = (9 * (int)cellcolor.red) / 10 + (1 * (int)bgcolor.red) / 10;
									blendcolor.green = (9 * (int)cellcolor.green) / 10 + (1 * (int)bgcolor.green) / 10;
									blendcolor.blue = (9 * (int)cellcolor.blue) / 10 + (1 * (int)bgcolor.blue) / 10;
								}
							}
						} else if (h == 0 || h == 4) { // 0 or 4
							if (w == 1 || w == 3) {
								blendcolor.red = (5 * (int)cellcolor.red) / 10 + (5 * (int)bgcolor.red) / 10;
								blendcolor.green = (5 * (int)cellcolor.green) / 10 + (5 * (int)bgcolor.green) / 10;
								blendcolor.blue = (5 * (int)cellcolor.blue) / 10 + (5 * (int)bgcolor.blue) / 10;
							} else if (w == 2) { // 2
								blendcolor.red = (9 * (int)cellcolor.red) / 10 + (1 * (int)bgcolor.red) / 10;
								blendcolor.green = (9 * (int)cellcolor.green) / 10 + (1 * (int)bgcolor.green) / 10;
								blendcolor.blue = (9 * (int)cellcolor.blue) / 10 + (1 * (int)bgcolor.blue) / 10;
							} else if (w == 0 || w == 4) { // 0 or 4
								blendcolor = bgcolor;
							}
						}*/

						image[height_idx_start + h][width_idx_start + w] = blendcolor;
					}
			} else if (temp.mode == 1) {
				if (temp.s == 0) { // apply to compareafferents
					compareafferents[lastafferent].x = temp.x;
					compareafferents[lastafferent].y = temp.y;
					compareafferents[lastafferent].z = temp.z;
					lastafferent++;

					// now render afferent node
					u32 height_idx_start = (temp.y * 2) * PX_PER_CELL;
					u32 width_idx_start = (temp.x * 2) * PX_PER_CELL;

					cellcolor.red = 180;
					cellcolor.green = 180;
					cellcolor.blue = 180;

					for (int h = 0; h < PX_PER_CELL*2; h++)
						for (int w = 0; w < PX_PER_CELL*2; w++) {
							blendcolor = cellcolor;

							if (h == 1 || h == PX_PER_CELL * 2 - 2 || w == 1 || w == PX_PER_CELL * 2 - 1) {
								blendcolor.red = 120;
								blendcolor.green = 120;
								blendcolor.blue = 120;
							}
							if ((h == 1 || h == PX_PER_CELL * 2 - 2) && (w == 1 || w == PX_PER_CELL * 2 - 1))
								blendcolor = bgcolor;
							if (h == 0 || h == PX_PER_CELL * 2 - 1 || w == 0)
								blendcolor = bgcolor;

							image[height_idx_start + h][width_idx_start + w] = blendcolor;
						}
				} else if (temp.s == 1) { // apply to comparehevs
					comparehevs[lasthev].x = temp.x;
					comparehevs[lasthev].y = temp.y;
					comparehevs[lasthev].z = temp.z;
					lasthev++;

					// now render afferent node
					u32 height_idx_start = (temp.y * 2) * PX_PER_CELL;
					u32 width_idx_start = (temp.x * 2) * PX_PER_CELL;

					cellcolor.red = 255;
					cellcolor.green = 236;
					cellcolor.blue = 139;

					for (int h = 0; h < PX_PER_CELL*2; h++)
						for (int w = 0; w < PX_PER_CELL*2; w++) {
							blendcolor = cellcolor;

							if (h == 1 || h == PX_PER_CELL * 2 - 2 || w == 1 || w == PX_PER_CELL * 2 - 1) {
								blendcolor.red = 200;
								blendcolor.green = 170;
								blendcolor.blue = 80;
							}
							if ((h == 1 || h == PX_PER_CELL * 2 - 2) && (w == 1 || w == PX_PER_CELL * 2 - 1))
								blendcolor = bgcolor;
							if (h == 0 || h == PX_PER_CELL * 2 - 1 || w == 0)
								blendcolor = bgcolor;

							image[height_idx_start + h][width_idx_start + w] = blendcolor;
						}
				}
			} else if (temp.mode == 2) { // apply to compareefferents
				compareefferents[lastefferent].x = temp.x;
				compareefferents[lastefferent].y = temp.y;
				compareefferents[lastefferent].z = temp.z;
				lastefferent++;
				//cout << "Found efferent: " << temp.x << ", " << temp.y << ", " << temp.z << endl;

				// now render efferent node
				u32 height_idx_start = (temp.y * 2) * PX_PER_CELL;
				u32 width_idx_start = (temp.x * 2) * PX_PER_CELL;

				cellcolor.red = 10;
				cellcolor.green = 210;
				cellcolor.blue = 10;

				for (int h = 0; h < PX_PER_CELL*2; h++)
					for (int w = 0; w < PX_PER_CELL*2; w++) {
						blendcolor = cellcolor;

						if (h == 1 || h == PX_PER_CELL * 2 - 2 || w == 1 || w == PX_PER_CELL * 2 - 1) {
							blendcolor.red = 0;
							blendcolor.green = 120;
							blendcolor.blue = 0;
						}
						if ((h == 1 || h == PX_PER_CELL * 2 - 2) && (w == 1 || w == PX_PER_CELL * 2 - 1))
							blendcolor = bgcolor;
						if (h == 0 || h == PX_PER_CELL * 2 - 1 || w == 0)
							blendcolor = bgcolor;

						image[height_idx_start + h][width_idx_start + w] = blendcolor;
					}
			}

			if (DISPLAY_SDL || WRITE_VIDEO) {
				SDL_PollEvent(&event);
				switch(event.type) {
					case SDL_QUIT:
						SDL_Quit();
						exit(0);
						break;
					default:
						break;
				}
			}
		} while (temp.mode != 3 && !infile.eof()); // until end-of-frame

		if (temp.mode == 3 && !infile.eof()) { // process end of frame
			framenumber = temp.x | (temp.y << 8) | (temp.z << 13) | (temp.s << 18);
			cout << "Frame " << framenumber << " processing - " << framenumber / 600 << " hrs, "
			<< cellcounts[CELL_H] << " HT cells, " << cellcounts[CELL_K] << " KT cells" << endl;

			if (WRITE_NUMCELLS) {
				numcellsfile << (float)framenumber / 600.0f << ",";
				for (int p = 0; p < NUMCELLTYPES; p++)
					numcellsfile << cellcounts[p] << ",";
				numcellsfile << lastafferent << "," << lastefferent << endl;
			}

			// now draw frame and output
			char file_name[100];
			FILE *fp;
			static u32 filecount = 0;
			png_structp png_ptr;
			png_infop info_ptr;
			if (WRITE_PNG) {
				sprintf(file_name, "images/img%d.png", filecount++);
				fp = fopen(file_name, "wb");
				globalfd = fp;
				if (!fp) {
					cout << "Couldn't open PNG" << endl;
					exit(1);
				}
				// width = 200 -> 5 px per = 1000 px
				// height = 25 * 4 -> 5 px per = 500 px

				png_ptr = png_create_write_struct
					(PNG_LIBPNG_VER_STRING, 0, 0, 0);
				if (!png_ptr) {
					cout << "Unable to create write struct" << endl;
					exit(1);
				}

				info_ptr = png_create_info_struct(png_ptr);
				if (!info_ptr) {
					png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
					cout << "Unable to create info struct" << endl;
					exit(1);
				}

				if (setjmp(png_jmpbuf(png_ptr))) {
				   png_destroy_write_struct(&png_ptr, &info_ptr);
				   fclose(fp);
				   cout << "HUGE ERROR!" << endl;
				   exit(1);
				}

#ifdef WIN32
				png_set_write_fn(png_ptr, fp, (png_rw_ptr)rbt_fwrite, (png_flush_ptr)rbt_fflush);
#else
				png_init_io(png_ptr, fp);
#endif
				png_set_IHDR(png_ptr, info_ptr, IMAGE_WIDTH, IMAGE_HEIGHT,
				   8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
				   PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
				png_set_sRGB(png_ptr, info_ptr, PNG_sRGB_INTENT_PERCEPTUAL);

				png_write_info(png_ptr, info_ptr);
		    }

			// update image
/*			for (int x = 0; x < GRID_X; x++)
				for (int y = 0; y < GRID_Y; y++)
					for (int z = 0; z < GRID_Z; z++)
						for (int s = 0; s < 4; s++) {
							u32 height_idx_start = (y * 4 + s) * PX_PER_CELL;
							u32 height_idx_end = height_idx_start + PX_PER_CELL - 1;
							u32 width_idx_start = x * PX_PER_CELL;
							u32 width_idx_end = width_idx_start + PX_PER_CELL - 1;

							png_color cellcolor, blendcolor, bgcolor, whitecolor;

							bgcolor.red = bgcolor.green = bgcolor.blue = 0;
							whitecolor.red = whitecolor.green = whitecolor.blue = 255;

							if (comparegrid[x][y][z][s].type == CELL_H) {
								cellcolor.red = 0;
								cellcolor.green = 180;
								cellcolor.blue = 255;
							} else if (comparegrid[x][y][z][s].type == CELL_K) {
								cellcolor.red = 255;
								cellcolor.green = 0;
								cellcolor.blue = 0;
							} else if (comparegrid[x][y][z][s].type == CELL_D) {
								cellcolor.red = 0;
								cellcolor.green = 255;
								cellcolor.blue = 0;
							} else {
								cellcolor = bgcolor;
							}

							for (int h = 0; h < PX_PER_CELL; h++)
								for (int w = 0; w < PX_PER_CELL; w++) {
									//blendcolor = whitecolor;
									blendcolor = bgcolor;
									if (h == 1 || h == 2 || h == 3 || h == 4 || h == 5) {
										if (w == 1 || w == 2 || w == 3 || w == 4 || w == 5) {
											blendcolor = cellcolor;
										}
									}
									if ((h == 1 || h == 5) && (w == 1 || w == 5)) {
										blendcolor.red = (5 * (int)cellcolor.red) / 10 + (5 * (int)bgcolor.red) / 10;
										blendcolor.green = (5 * (int)cellcolor.green) / 10 + (5 * (int)bgcolor.green) / 10;
										blendcolor.blue = (5 * (int)cellcolor.blue) / 10 + (5 * (int)bgcolor.blue) / 10;
										//blendcolor = bgcolor;
									}

									/*if (h == 1 || h == 2 || h == 3 || h == 4 || h == 5) {
										if (w == 1 || w == 2 || w == 3 || w == 4 || w == 5) {
											blendcolor = cellcolor;
										} else if (h == 1 || h == 4) { // 0.3
											blendcolor.red = (3 * (int)cellcolor.red) / 10 + (7 * (int)bgcolor.red) / 10;
											blendcolor.green = (3 * (int)cellcolor.green) / 10 + (7 * (int)bgcolor.green) / 10;
											blendcolor.blue = (3 * (int)cellcolor.blue) / 10 + (7 * (int)bgcolor.blue) / 10;
										} else if (h == 2 || h == 3) { // 0.8
											blendcolor.red = (8 * (int)cellcolor.red) / 10 + (2 * (int)bgcolor.red) / 10;
											blendcolor.green = (8 * (int)cellcolor.green) / 10 + (2 * (int)bgcolor.green) / 10;
											blendcolor.blue = (8 * (int)cellcolor.blue) / 10 + (2 * (int)bgcolor.blue) / 10;
										}
									} else {
										if (w == 0 || w == 5) {
											blendcolor = bgcolor;
										} else if (w == 1 || w == 4) { // 0.3
											blendcolor.red = (3 * (int)cellcolor.red) / 10 + (7 * (int)bgcolor.red) / 10;
											blendcolor.green = (3 * (int)cellcolor.green) / 10 + (7 * (int)bgcolor.green) / 10;
											blendcolor.blue = (3 * (int)cellcolor.blue) / 10 + (7 * (int)bgcolor.blue) / 10;
										} else if (w == 2 || w == 3) { // 0.8
											blendcolor.red = (8 * (int)cellcolor.red) / 10 + (2 * (int)bgcolor.red) / 10;
											blendcolor.green = (8 * (int)cellcolor.green) / 10 + (2 * (int)bgcolor.green) / 10;
											blendcolor.blue = (8 * (int)cellcolor.blue) / 10 + (2 * (int)bgcolor.blue) / 10;
										}
									}*/

									// FOR PX_PER_CELL = 5
									/*if (h == 1 || h == 2 || h == 3) {
										if (w == 1 || w == 2 || w == 3) {
											blendcolor = cellcolor;
										} else if (w == 0 || w == 4) {
											if (h == 1 || h == 3) {
												blendcolor.red = (5 * (int)cellcolor.red) / 10 + (5 * (int)bgcolor.red) / 10;
												blendcolor.green = (5 * (int)cellcolor.green) / 10 + (5 * (int)bgcolor.green) / 10;
												blendcolor.blue = (5 * (int)cellcolor.blue) / 10 + (5 * (int)bgcolor.blue) / 10;
											} else if (h == 2) { // 2
												blendcolor.red = (9 * (int)cellcolor.red) / 10 + (1 * (int)bgcolor.red) / 10;
												blendcolor.green = (9 * (int)cellcolor.green) / 10 + (1 * (int)bgcolor.green) / 10;
												blendcolor.blue = (9 * (int)cellcolor.blue) / 10 + (1 * (int)bgcolor.blue) / 10;
											}
										}
									} else if (h == 0 || h == 4) { // 0 or 4
										if (w == 1 || w == 3) {
											blendcolor.red = (5 * (int)cellcolor.red) / 10 + (5 * (int)bgcolor.red) / 10;
											blendcolor.green = (5 * (int)cellcolor.green) / 10 + (5 * (int)bgcolor.green) / 10;
											blendcolor.blue = (5 * (int)cellcolor.blue) / 10 + (5 * (int)bgcolor.blue) / 10;
										} else if (w == 2) { // 2
											blendcolor.red = (9 * (int)cellcolor.red) / 10 + (1 * (int)bgcolor.red) / 10;
											blendcolor.green = (9 * (int)cellcolor.green) / 10 + (1 * (int)bgcolor.green) / 10;
											blendcolor.blue = (9 * (int)cellcolor.blue) / 10 + (1 * (int)bgcolor.blue) / 10;
										} else if (w == 0 || w == 4) { // 0 or 4
											blendcolor = bgcolor;
										}
									}*/
/*
									image[height_idx_start + h][width_idx_start + w] = blendcolor;
								}
						}
*/
			if (WRITE_PNG) {
			    png_write_image(png_ptr, row_pointers);
				png_write_end(png_ptr, info_ptr);

				png_destroy_write_struct(&png_ptr, &info_ptr);

				fclose(fp);
			}

			AVPicture pict;
			if (DISPLAY_SDL) {
				SDL_LockYUVOverlay(bmp);
				pict.data[0] = bmp->pixels[0];
				pict.data[1] = bmp->pixels[2];
				pict.data[2] = bmp->pixels[1];

				pict.linesize[0] = bmp->pitches[0];
				pict.linesize[1] = bmp->pitches[2];
				pict.linesize[2] = bmp->pitches[1];
				sws_scale(img_convert_ctx, pFrameRGB->data,
					pFrameRGB->linesize, 0, IMAGE_HEIGHT,
					pict.data, pict.linesize);
				SDL_UnlockYUVOverlay(bmp);

				rect.x = 0;
				rect.y = 0;
				rect.w = IMAGE_WIDTH;
				rect.h = IMAGE_HEIGHT;
				SDL_DisplayYUVOverlay(bmp, &rect);
			}

			static int x,y;
			i=0;
			if (WRITE_VIDEO) {
				sws_scale(video_convert_ctx, pFrameRGB->data,
					pFrameRGB->linesize, 0, IMAGE_HEIGHT,
					picture->data, picture->linesize);
		        /*for(y=0;y<c->height;y++) {
		            for(x=0;x<c->width;x++) {
		                picture->data[0][y * picture->linesize[0] + x] = x + y + i * 3;
		            }
		        }

		        // Cb and Cr
		        for(y=0;y<c->height/2;y++) {
		            for(x=0;x<c->width/2;x++) {
		                picture->data[1][y * picture->linesize[1] + x] = 128 + y + i * 2;
		                picture->data[2][y * picture->linesize[2] + x] = 64 + x + i * 5;
		            }
		        }*/

		        // encode the image
		        out_size = avcodec_encode_video(c, outbuf, outbuf_size, picture);
		        printf("encoding frame %3d (size=%5d)\n", i, out_size);
		        fwrite(outbuf, 1, out_size, f);
		        i++;
			}
		}

		if (DISPLAY_SDL || WRITE_VIDEO) {
			SDL_PollEvent(&event);
			switch(event.type) {
				case SDL_QUIT:
					SDL_Quit();
					exit(0);
					break;
				default:
					break;
			}
		}
	}

	if (WRITE_VIDEO) {
		/* get the delayed frames */
	    for(; out_size; i++) {
	        fflush(stdout);

	        out_size = avcodec_encode_video(c, outbuf, outbuf_size, NULL);
	        printf("write frame %3d (size=%5d)\n", i, out_size);
	        fwrite(outbuf, 1, out_size, f);
	    }

	    /* add sequence end code to have a real mpeg file */
	    outbuf[0] = 0x00;
	    outbuf[1] = 0x00;
	    outbuf[2] = 0x01;
	    outbuf[3] = 0xb7;
	    fwrite(outbuf, 1, 4, f);
	    fclose(f);
	    free(picture_buf);
	    free(outbuf);

	    avcodec_close(c);
	    av_free(c);
	    av_free(picture);
	}

	if (WRITE_NUMCELLS)
		numcellsfile.close();
	infile.close();

	return 0;
}


// Replaces gluPerspective. Sets the frustum to perspective mode.
// fovY     - Field of vision in degrees in the y direction
// aspect   - Aspect ratio of the viewport
// zNear    - The near clipping distance
// zFar     - The far clipping distance
void perspectiveGL(GLdouble fovY, GLdouble aspect, GLdouble zNear, GLdouble zFar)
{
	GLdouble fW, fH;

	fH = tan( (fovY / 2) / 180 * M_PI ) * zNear;
	fW = fH * aspect;

	glFrustum( -fW, fW, -fH, fH, zNear, zFar );
}


/* Clamp a value that is not between min and max */
GLfloat clamp(GLfloat value, GLfloat min, GLfloat max)
{
	if (value > max)
		return max;
	else if (value < min)
		return min;
	else
		return value;
}


