#include "stdafx.h"
using namespace std;

/*#ifdef WIN32
FILE * globalfd;
size_t rbt_fwrite(int fd, const void *buf, size_t count) {
	return fwrite(buf, 1, count, globalfd); }
int rbt_fflush(FILE * stream) {	return fflush(stream); }
#endif*/

void applyvector(float origx, float origy, float origz, float &tempx, float &tempy, float &tempz);

float zoffset = -100.0f, xoffset = 0.0, yoffset = 0.0, xrotation = 0.0, yrotation = 0.0; // drawing parameters, 5 axis
unsigned char domove = 2;

urandom rnd(Rand_Sfmt);

#ifdef WIN32
#warning "WIN32 - using _tmain"
int _tmain(int argc, _TCHAR* argv[]) {
#else
#warning "no WIN32 - using main"
int main (int argc, char *argv[]) {
#endif
//	assert(sizeof(OutStruct) == 4);
//	assert(sizeof(GridCoord) == 3);
//	assert(sizeof(CompareStruct) == 1);	
	

#ifdef GR_MODEL
	printf("GR element: %u; Array: %u MB\n", sizeof(GRElement), sizeof(GRElement)*GR_GRID_X*GR_GRID_Y*GR_GRID_Z/(1024*1024));
	printf("GR diffuse: %u; Array: %u MB\n", sizeof(GRDiffuse), sizeof(GRDiffuse)*GR_DIF_X*GR_DIF_Y*GR_DIF_Z/(1024*1024));
	GRGrid * grg = new GRGrid();
	VizGrid * vg = grg;
#else
	LNGrid * lng = new LNGrid();
	VizGrid * vg = lng;
#endif

//	srand(time(NULL));
	u32 seedval = time(NULL) ^ clock();
	rnd.settype(Rand_Sfmt, seedval);
	cout << "Seed: " << rnd.getseed() << endl;

	vg->InitStructure();

	int frame=0;
	int done;
	int slice = 0;
	for (done = 0; !done;) {
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glLoadIdentity();
		glRotatef(yrotation, 1, 0, 0);
		glRotatef(xrotation, 0, 1, 0);
		glTranslatef(xoffset, yoffset, zoffset);

		if (domove >= 1) {
#ifdef GR_MODEL
			u32 startclk = clock();
			for (int p = 0; p < 10; p++) { // run 6 minutes = 0.1h
				vg->CalcTimeSlice(slice); // run every 6 seconds, T cells every 18 seconds; dcs will run every 72 seconds
				if (domove == 1) {
					domove = 0;
					break;
				}
			//	slice += 3; // was 20 = 120s, now 3 = 18s
				slice++;
			}
			printf("Simulation rate: %0.2f min/min\n", ((float)(clock() - startclk)/(float)CLOCKS_PER_SEC) / 60.0f);

#else
			for (int p = 0; p < 30; p++) { // run 1 hour
				vg->CalcTimeSlice(slice); // run every 120 seconds; dcs will run every 72 seconds
				if (domove == 1) {
					domove = 0;
					break;
				}
				slice += 20; // 20 = 120s
			}
#endif
		}
		vg->DrawGrid();
		
		SDL_GL_SwapBuffers();

#if defined(WRITE_PNG) || defined(WRITE_VIDEO)
		vg->ExtractFrame();
#endif

		// now draw frame and output
#ifdef WRITE_PNG
		char file_name[100];
		FILE *fp;
		static u32 filecount = 0;
		png_structp png_ptr;
		png_infop info_ptr;

		sprintf(file_name, "images/img%d.png", filecount++);
		fp = fopen(file_name, "wb");
		if (!fp) {
			cout << "Couldn't open PNG" << endl;
			exit(1);
		}
		// width = 200 -> 5 px per = 1000 px
		// height = 25 * 4 -> 5 px per = 500 px

		png_ptr = png_create_write_struct
			(PNG_LIBPNG_VER_STRING, 0, 0, 0);
		if (!png_ptr) {
			cout << "Unable to create write struct" << endl;
			exit(1);
		}

		info_ptr = png_create_info_struct(png_ptr);
		if (!info_ptr) {
			png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
			cout << "Unable to create info struct" << endl;
			exit(1);
		}

		if (setjmp(png_jmpbuf(png_ptr))) {
		   png_destroy_write_struct(&png_ptr, &info_ptr);
		   fclose(fp);
		   cout << "HUGE ERROR!" << endl;
		   exit(1);
		}

		png_init_io(png_ptr, fp);

		png_set_IHDR(png_ptr, info_ptr, FRAME_WIDTH, FRAME_HEIGHT,
		   8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
		   PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
		png_set_sRGB(png_ptr, info_ptr, PNG_sRGB_INTENT_PERCEPTUAL);

		png_write_info(png_ptr, info_ptr);
		
	    png_write_image(png_ptr, row_pointers);
		png_write_end(png_ptr, info_ptr);

		png_destroy_write_struct(&png_ptr, &info_ptr);

		fclose(fp);
#endif

#ifdef WRITE_VIDEO
		vg->WriteVideo();
#endif

		SDL_Event event;
		while (SDL_PollEvent(&event)) {
			float origx = 0.0, origy = 0.0, origz = 0.0;
			float tempx = 0.0, tempy = 0.0, tempz = 0.0;
			if (event.type == SDL_KEYDOWN) {
				if (event.key.keysym.sym == SDLK_ESCAPE) {
					done = 1;
				} else if (event.key.keysym.sym == SDLK_f && event.key.state == SDL_PRESSED) {
					extern bool ENABLE_FRC;
					ENABLE_FRC = !ENABLE_FRC;
				} else if (event.key.keysym.sym == SDLK_o && event.key.state == SDL_PRESSED) {
					extern bool ENABLE_OUTLINE, ENABLE_GROUTLINE;
					ENABLE_OUTLINE = !ENABLE_OUTLINE;
					ENABLE_GROUTLINE = !ENABLE_GROUTLINE;
				} else if (event.key.keysym.sym == SDLK_w && event.key.state == SDL_PRESSED) {
					// [0 -2 0] is vector
					origx = 0.0; origy = 0.0; origz = 5.0; // y = -2
					applyvector(origx, origy, origz, tempx, tempy, tempz);
					xoffset += tempx; yoffset += tempy; zoffset += tempz;
				} else if (event.key.keysym.sym == SDLK_s && event.key.state == SDL_PRESSED) {
					origx = 0.0; origy = 0.0; origz = -5.0; // y = 2
					applyvector(origx, origy, origz, tempx, tempy, tempz);
					xoffset += tempx; yoffset += tempy; zoffset += tempz;
				} else if (event.key.keysym.sym == SDLK_a && event.key.state == SDL_PRESSED) {
					origx = 10.0; origy = 0.0; origz = 0.0;
					applyvector(origx, origy, origz, tempx, tempy, tempz);
					xoffset += tempx; yoffset += tempy; zoffset += tempz;
				} else if (event.key.keysym.sym == SDLK_d && event.key.state == SDL_PRESSED) {
					origx = -10.0; origy = 0.0; origz = 0.0;
					applyvector(origx, origy, origz, tempx, tempy, tempz);
					xoffset += tempx; yoffset += tempy; zoffset += tempz;
				} else if (event.key.keysym.sym == SDLK_COMMA && event.key.state == SDL_PRESSED) {
					domove = 0;
				} else if (event.key.keysym.sym == SDLK_PERIOD && event.key.state == SDL_PRESSED) {
					domove = 1;
				} else if (event.key.keysym.sym == SDLK_SLASH && event.key.state == SDL_PRESSED) {
					domove = 2;
				}
			} else if (event.type == SDL_MOUSEBUTTONDOWN) {
				if (event.button.button == SDL_BUTTON_WHEELUP) {
					origx = 0.0; origy = 0.0; origz = 5.0;
					applyvector(origx, origy, origz, tempx, tempy, tempz);
					xoffset += tempx; yoffset += tempy; zoffset += tempz;
				} else if (event.button.button == SDL_BUTTON_WHEELDOWN) {
					origx = 0.0; origy = 0.0; origz = -5.0;
					applyvector(origx, origy, origz, tempx, tempy, tempz);
					xoffset += tempx; yoffset += tempy; zoffset += tempz;
				}
			} else if (event.type == SDL_MOUSEMOTION) {
				if (event.motion.state == 0x4) { // right button only
					xrotation += event.motion.xrel;
					yrotation += event.motion.yrel;
				}
				if (event.motion.state == 0x1) { // left button only
					origx = (float)event.motion.xrel/5.0f; origy = -(float)event.motion.yrel/5.0f; origz = 0.0; // event.motion.yrel
					applyvector(origx, origy, origz, tempx, tempy, tempz);
					xoffset += tempx; yoffset += tempy; zoffset += tempz;
				}
			}
		}
	}

#ifdef WRITE_VIDEO
	vg->FinalizeVideo();
#endif

/*
	if (WRITE_NUMCELLS) {
		//enum CellTypes { CELL_NONE = 0, CELL_ALL = 0, CELL_H = 1, CELL_K = 2, CELL_D = 3,
		//CELL_TG = 4, CELL_TC = 5, CELL_TR = 6, CELL_M = 7 }; // 8
		numcellsfile.open("numcells.csv");//, ios::out | out::trunc); // weird error with ios::out
		numcellsfile << "Time (hrs),# None Cells,# Helper T Cells,# Killer T Cells,# Dendritic Cells," <<
			"# TG Cells,# TC Cells,# TR Cells,# Macrophages,# Afferent Nodes,# Efferent Nodes" << endl;
	}
			if (WRITE_NUMCELLS) {
				numcellsfile << (float)framenumber / 600.0f << ",";
				for (int p = 0; p < NUMCELLTYPES; p++)
					numcellsfile << cellcounts[p] << ",";
				numcellsfile << lastafferent << "," << lastefferent << endl;
			}
*/
	return 0;
}

void applyvector(float origx, float origy, float origz, float &tempx, float &tempy, float &tempz) {
	tempx = origx;
	tempy = cos(yrotation*(M_PI/180.0))*origy + sin(yrotation*(M_PI/180.0))*origz;
	tempz = -sin(yrotation*(M_PI/180.0))*origy + cos(yrotation*(M_PI/180.0))*origz;

	origx = tempx; origy = tempy; origz = tempz;

	tempx = cos(xrotation*(M_PI/180.0))*origx - sin(xrotation*(M_PI/180.0))*origz;
	tempy = origy;
	tempz = sin(xrotation*(M_PI/180.0))*origx + cos(xrotation*(M_PI/180.0))*origz;
}
