#include "stdafx.h"
#include "Element.h"

#ifndef __VIZGRID_H__
#define __VIZGRID_H__

//#define DRAWPOINTS 1
//#define SHOWALLCELLS	1
#ifdef GR_MODEL
#define DISPGRIDBOX	1
//#define DISPGRIDY	1
//#define DISPGRIDZ	1
#else
//#define DISPGRIDBOX	1
#define DISPGRIDY	1
//#define DISPGRIDZ	1
#endif
#define INITIALFILL	1

#define WRITE_VIDEO		1
//#define WRITE_PNG		1
//#define WRITE_NUMCELLS	0

#define FRAME_WIDTH		1024
#define FRAME_HEIGHT	768
//#define FRAME_WIDTH		640
//#define FRAME_HEIGHT	480
#define SPHERE_RES		20
#define VIDEO_FMT		PIX_FMT_YUV420P

#define LN_GRID_X	400
#define LN_GRID_Y	50
// was 50 high, make 100?
#define LN_GRID_Z	40

#define GR_GRID_X	400
#define GR_GRID_Y	400
#define GR_GRID_Z	400


#include "3dmovement.h"

// Disable Windows specifc code, while trying to get the code to run under Linux.
#ifdef CLOCK
GLvoid glPrint(const char *fmt, ...);
#endif

/*enum textquality {solid, shaded, blended};
SDL_Surface *drawtext(TTF_Font *fonttodraw, char fgR, char fgG, char fgB, char fgA, 
	char bgR, char bgG, char bgB, char bgA, char * text, textquality quality);*/

extern urandom rnd;
class VizGrid {
	public:
		VizGrid();
		
		virtual void InitStructure() {}
		virtual bool CalcTimeSlice(int slice) {return false;} // return true if any changes made
		virtual void DrawGrid() {}

		void ExtractFrame();
		void WriteVideo();
		void FinalizeVideo();

	protected:
		void InitSDL();
		void InitGL();
		void InitPNG();
		void InitVideo();

		coord applydir(coord current, int newdir);
		int getnumdirs(int dir);

		// SDL
		SDL_Surface *screen;

		// TTF
		TTF_Font * timerfont;
		SDL_Surface *fontsurface;

		// GL
		GLUquadric *quadric;
		GLUquadricObj *sphere;
		GLuint texture;
		GLuint mysphereID, mysphereIDbig, mycylID;

		// PNG
		png_color image[FRAME_HEIGHT][FRAME_WIDTH];
		png_byte *row_pointers[FRAME_HEIGHT];

		// Video
		unsigned int frame;
		AVFrame * pFrameRGB;
		struct SwsContext *img_convert_ctx, *video_convert_ctx;
		AVCodec *codec;
		AVCodecContext *c;
		AVFrame *picture;
		FILE *f;
		int outbuf_size;
		uint8_t *outbuf, *picture_buf;
		
		// simulation control
		int timeindex; // 6 second units, 20 = 2 minutes
		int totalgridtime; // to figure out how long cells stay on the grid
		int totalcellslost;
};

class LNGrid : public VizGrid {
	public:
		void InitStructure();
		bool CalcTimeSlice(int slice); // return true if any changes made
		void Recruitment();
		void Movement();
		void ClearMoved();
		void States();
		void DrawGrid();
		void ClearDead();

	private:
		int FRC_generator(int num_iterations);
		bool InMoore(coordc currloc, CellTypes tofind, coord * loc);
		bool InMooreState(coordc currloc, CellTypes tofind, CellStates tomatch, coord * loc);
		bool InMooreStateRadius(coordc currloc, CellTypes tofind, CellStates tomatch, float radius, coord * loc);

		LNElement grid[LN_GRID_X][LN_GRID_Y][LN_GRID_Z];

		vector<coord> als, els, hevs;
		int totalcells[NUMCELLTYPES];
		double mfpsum, nummfp;
};

/* DIFFUSION SECTION
	There is an upper-limit for dt: dt<(dx^2)/(6*D), where D is the largest diffusion 
	coefficient used for updating diffusion on a grid with microcompartment size of dx. 
	Since, Dmax used in our simulation will be 10-7 cm2/s -> with dx = 10um, we need 
	dt<1.66 sec! However, if we update diffusion on the grid in a more corse-grained 
	manner, say for each 20um*20um*20um at each time step, we need dt<6.4 sec. Thus, 
	dt=6 works. But, we might face problems with accuracy in chemotactic movement of 
	T cells.
	(20^2 um^2)/(60 um^2/s) = 6.67s = 1.11t
	(40^2 um^2)/(60 um^2/s) = 26.67s = 4.44t
	(80^2 um^2)/(60 um^2/s) = 106.67s = 17.78t
	*/

#define DECIMATE_DIFFUSION	2
#define DIFFUSION_TIMESTEP	1
//#define DECIMATE_DIFFUSION	4
//#define DIFFUSION_TIMESTEP	4
//#define DECIMATE_DIFFUSION	8
//#define DIFFUSION_TIMESTEP	18
#define GR_DIF_X	(GR_GRID_X/DECIMATE_DIFFUSION + 2)
#define GR_DIF_Y	(GR_GRID_Y/DECIMATE_DIFFUSION + 2)
#define GR_DIF_Z	(GR_GRID_Z/DECIMATE_DIFFUSION + 2)
#define DX_DIFFUSION	(5 * DECIMATE_DIFFUSION)

#define D_TNF			((10e8) * (6.0f * TNF_DIFFUSION_COEFF * DIFFUSION_TIMESTEP) / (DX_DIFFUSION * DX_DIFFUSION))
#define deg_TNF			(DEG_TNF * DIFFUSION_TIMESTEP)
#define D_CCL2			((10e8) * (6.0f * CCL2_DIFFUSION_COEFF * DIFFUSION_TIMESTEP) / (DX_DIFFUSION * DX_DIFFUSION))
#define deg_CCL2		(DEG_CCL2 * DIFFUSION_TIMESTEP)
#define D_CCL5			((10e8) * (6.0f * CCL5_DIFFUSION_COEFF * DIFFUSION_TIMESTEP) / (DX_DIFFUSION * DX_DIFFUSION))
#define deg_CCL5		(DEG_CCL5 * DIFFUSION_TIMESTEP)
#define D_CXCL9			((10e8) * (6.0f * CXCL9_DIFFUSION_COEFF * DIFFUSION_TIMESTEP) / (DX_DIFFUSION * DX_DIFFUSION))
#define deg_CXCL9		(DEG_CXCL9 * DIFFUSION_TIMESTEP)
#define D_EXTMTB		((10e8) * (6.0f * EXTMTB_DIFFUSION_COEFF * DIFFUSION_TIMESTEP) / (DX_DIFFUSION * DX_DIFFUSION))
#define deg_EXTMTB		(DEG_EXTMTB * DIFFUSION_TIMESTEP)

class GRGrid : public VizGrid {
	public:
		void InitStructure();
		bool CalcTimeSlice(int slice); // return true if any changes made
		void CalcDiffusion(int dt);
		void Recruitment();
		void Movement();
		void ClearMoved();
		void States();
		void DrawGrid();
		void ClearDead();

	private:
		bool InMoore(coordc currloc, CellTypes tofind, coord * loc);
		bool InMooreState(coordc currloc, CellTypes tofind, CellStates tomatch, coord * loc);
		bool InMooreStateRadius(coordc currloc, CellTypes tofind, CellStates tomatch, float radius, coord * loc);
		coordc GenEmptyCoord(coordc currloc, int delta);
		int ClearSphere(coordc currloc, float radius);

		DIFFUSETYPE bilinear_3d(coordc currloc, ChemTypes toprocess);

		GRElement grid[GR_GRID_X][GR_GRID_Y][GR_GRID_Z];
		GRDiffuse chem[GR_DIF_X][GR_DIF_Y][GR_DIF_Z];

		int totalcells[NUMCELLTYPES];
};

#endif
