#ifndef __CRITICAL_PARAMS_H__
#define __CRITICAL_PARAMS_H__


struct coord {
	int x, y, z;
};

class coordc {
	public:
		coordc(int a, int b, int c) { x = a; y = b; z = c; }
		int x, y, z;
};

// Movement properties
/*#define WALL_BOUNCE		1
#define FOCUSED_MOVE	0

// Moore Neighborhood Size
#if (GRID_Z == 1)
#define NEIGHBORS 9
#define SLOTS 4
#else
#define NEIGHBORS 27
#define SLOTS 8
#endif

// ***** Lymph Node Model Constants *****
// 40, 100, 160 originally
#define NUM_HEV			30
#define NUM_AFFERENT	20
#define NUM_EFFERENT	60
#define MAX_HT			262144
#define MAX_KT			262144
#define MAX_DC			262144
#define TIMESTEP		20
*/
// Timing
#define MINUTE			10
#define HOUR			(60 * MINUTE)
#define DAY				(24 * HOUR)
//#define pMHC_HALFLIFE	60
//#define MAX_DCS			300

// CD4	
#define CD4_RESTING_MOVEMENT_PERIOD			(2 * MINUTE)
#define CD4_ACTIVE_MOVEMENT_PERIOD			(4 * MINUTE)
#define CD4_MIN_AGE							(165 * DAY)
#define CD4_MAX_AGE							(365 * DAY)
#define CD4_MAX_AGE_STD						(15 * DAY)
#define CD4_pMHC_FOR_MEDIAN_BINDING_PROB	150
#define CD4_SHAPE_BINDING_PROB_CURVE		15
#define PRIMING_THRESHOLD_VALUE				200
#define CD4_MAX_BINDING_TIME				(14 * HOUR)
#define CD4_MAX_BINDING_TIME_STD			(3 * HOUR)
#define UNBINDING_THRESHOLD					100
#define CD4_TIME_FOR_MEDIAN_PRIMING_PROB	((6 * HOUR) * 150)
#define CD4_SHAPE_PRIMING_PROB_CURVE		(CD4_TIME_FOR_MEDIAN_PRIMING_PROB / 10)
#define CD4_MAX_DIVISION					4
#define CD4_DOUBLING_PERIOD					(8 * HOUR)
#define CD4_MAX_TIME_EFFECTOR				(60 * HOUR)
#define CD4_MAX_TIME_EFFECTOR_STD			(1 * DAY)

// CD8
#define CD8_RESTING_MOVEMENT_PERIOD			(2 * MINUTE)
#define CD8_ACTIVE_MOVEMENT_PERIOD			(4 * MINUTE)
#define CD8_MIN_AGE							(165 * DAY)
#define CD8_MAX_AGE							(365 * DAY)
#define CD8_MAX_AGE_STD						(15 * DAY)
#define CD8_pMHC_FOR_MEDIAN_BINDING_PROB	150
#define CD8_SHAPE_BINDING_PROB_CURVE		15
#define CD8_MAX_BINDING_TIME				(14 * HOUR)
#define CD8_MAX_BINDING_TIME_STD			(3 * HOUR)
#define UNBINDING_THRESHOLD					100
#define CD8_TIME_FOR_MEDIAN_PRIMING_PROB	((6 * HOUR) * 150)
#define CD8_SHAPE_PRIMING_PROB_CURVE		(CD8_TIME_FOR_MEDIAN_PRIMING_PROB / 10)
#define CD8_MAX_DIVISION					8
#define CD8_DOUBLING_PERIOD					(8 * HOUR)
#define CD8_MAX_TIME_EFFECTOR				(60 * HOUR)
#define CD8_MAX_TIME_EFFECTOR_STD			(1 * DAY)

// DC
#define DC_MIN_AGE					(1 * DAY)
#define DC_MAX_AGE					(11 * DAY)
#define DC_MAX_AGE_STD				(2 * DAY)
#define pMHC_LIMIT					300
#define pMHC_DECAY_FACTOR			0.999980746097010
#define IDC_MDC_BOUNDARY			50
#define BINDING_THRESHOLD			200
#define UNBINDING_THRESHOLD			100
#define DC_MAX_TIME_MATURE			(60 * HOUR)
#define DC_MAX_TIME_MATURE_STD		(5 * HOUR)
#define PROB_eHELPER_LICENSES_MDC	50
#define DC_TIME_BEING_LICENSED		(36 * HOUR)
#define DC_TIME_BEING_LICENSED_STD	(4 * HOUR)

// simulation biological (in)variables
#define P_COGNATE_KCELL			300
#define P_COGNATE_HCELL			300

#define P_HCELL_TIME_ACTIVE		28800
#define P_KCELL_TIME_ACTIVE		28800
#define P_MIN_H_CELL_AGE		2376000
#define P_MAX_H_CELL_AGE		5256000
#define P_MAX_TIME_H_EFFECTOR	36000
#define P_MAX_TIME_K_EFFECTOR	36000
#define P_MIN_K_CELL_AGE		2376000
#define P_MAX_K_CELL_AGE		5256000
#define P_MIN_D_CELL_AGE		14400
#define P_MAX_D_CELL_AGE		158400

#define P_PROB_H_RECRUITMENT	0.99
#define P_PROB_K_RECRUITMENT	0.95
//#define P_PROB_D_RECRUITMENT	0.50
#define P_PROB_D_RECRUITMENT	0.04
// DC 5%

// ***** Granuloma Model Constants *****

// (code = paper / 1000)
#define NUM_VSOURCES 	49 // Vascular Sources

// Cell constants
#define MAX_MP_AGE 		1440000 // 60*24*(100 days) * 10
#define INITIAL_N_MP 	105

#define MP_CCL5_MIN 	20
#define MP_CCL5_SAT 	5790
#define MP_CCL2_MIN 	(0.1 * MP_CCL5_MIN)
#define MP_CCL2_SAT 	(10 * MP_CCL5_SAT)

#define TC_CCL5_MIN 	(MP_CCL5_MIN)
#define TC_CCL5_SAT 	(MP_CCL5_SAT)
#define TC_CXCL9_MIN 	(MP_CCL5_MIN)
#define TC_CXCL9_SAT 	(10 * MP_CCL5_SAT)

#define TG_CCL2_MIN 	(0.1 * MP_CCL5_MIN)
#define TG_CCL2_SAT 	(10 * MP_CCL5_SAT)
#define TG_CCL5_MIN 	(MP_CCL5_MIN)
#define TG_CCL5_SAT 	(MP_CCL5_SAT)
#define TG_CXCL9_MIN 	(MP_CCL5_SAT)
#define TG_CXCL9_SAT 	(10 * MP_CCL5_SAT)

#define TR_CCL5_MIN 	(0.1 * MP_CCL5_MIN)
#define TR_CCL5_SAT 	(MP_CCL5_SAT)

#define TNF_DIFFUSION_COEFF		0.000000052
#define DEG_TNF					0.000458
#define CCL2_DIFFUSION_COEFF	0.000000052
#define DEG_CCL2				0.000458
#define CCL5_DIFFUSION_COEFF	0.000000052
#define DEG_CCL5				0.000458
#define CXCL9_DIFFUSION_COEFF	0.000000052
#define DEG_CXCL9				0.000458
#define EXTMTB_DIFFUSION_COEFF	0.000000052
#define DEG_EXTMTB				0.000458

// CD4
#define kTACE_CD4_TCELL			0.00044
#define MIN_TNFR1_CD4_TCELL		100
#define MAX_TNFR1_CD4_TCELL		1000
#define MIN_TNFR2_CD4_TCELL		100
#define MAX_TNFR2_CD4_TCELL		1000
#define MIN_CCL2_CD4_TCELL		20
#define SAT_CCL2_CD4_TCELL		20
#define MIN_CCL5_CD4_TCELL		20
#define SAT_CCL5_CD4_TCELL		5000
#define MIN_CXCL9_CD4_TCELL		5000
#define SAT_CXCL9_CD4_TCELL		5000

// CD8
#define kTACE_CD8_TCELL			0.00044
#define MIN_TNFR1_CD8_TCELL		100
#define MAX_TNFR1_CD8_TCELL		1000
#define MIN_TNFR2_CD8_TCELL		100
#define MAX_TNFR2_CD8_TCELL		1000
#define MIN_CCL5_CD8_TCELL		20
#define SAT_CCL5_CD8_TCELL		20
#define MIN_CXCL9_CD8_TCELL		5000
#define SAT_CXCL9_CD8_TCELL		5000
 
// TREG
#define kTACE_REG_TCELL			0.00044
#define MIN_TNFR1_REG_TCELL		100
#define MAX_TNFR1_REG_TCELL		1000
#define MIN_TNFR2_REG_TCELL		100
#define MAX_TNFR2_REG_TCELL		1000
#define MIN_CCL5_REG_TCELL		20
#define SAT_CCL5_REG_TCELL		5000

// M
#define kTACE_MAC				0.00044
#define MIN_TNFR1_MAC			500
#define MAX_TNFR1_MAC			2000
#define MIN_TNFR2_MAC			500
#define MAX_TNFR2_MAC			2000
#define MIN_CCL2_MAC			20
#define SAT_CCL2_MAC			5000
#define MIN_CCL5_MAC			20
#define SAT_CCL5_MAC			5000
#define ksynth_MAC				0.21
#define INCR_CCL2_MAC			3000
#define INCR_CCL5_MAC			3000
#define INCR_CXCL9_MAC			6000
#define N_EXT_Mtb_NFkB_ACTIVATION			50
#define PROB_GAMMA_IFN_MAC_ACTIVATION			0.01
#define N_BACTERIA_ENGULFED_MACROPHAGE			1 
#define PROB_INFECTION_RESTING_MACROPHAGE			0.1
#define PROB_CD4_MAC_FAS_APOPTOSIS			0.01
#define N_TIMES_BECOME_CASEOUS			10 
#define PROB_CTL_MAC_KILLING			0.01
#define N_BACTERIA_BECOME_CHRONIC			10 
#define INTRA_CELL_GROWTH_RATE			0.0015
#define PROB_CD4_MAC_FAS_APOPTOSIS			0.01
#define PROB_CTL_CINFECTED_CLEAN_KILLING			0.75
#define N_BACTERIA_BURSTING			20

// D
#define kTACE_DC			0.00044
#define MIN_TNFR1_DC		500
#define MAX_TNFR1_DC		5000
#define MIN_TNFR2_DC		500
#define MAX_TNFR2_DC		5000
#define MIN_CCL2_DC		20
#define SAT_CCL2_DC		5000
#define MIN_CCL5_DC		20
#define SAT_CCL5_DC		5000
#define PROB_UPTAKE_BACTERIA_DC		0.1
#define N_BACTERIA_ENGULFED_DC		1


#endif
