#include "stdafx.h"

static CRan32 cran32(0,0);
static BasicLCG lcg(0);

#ifdef USE_BOOST
static boost::mt19937 tr1mt(1);
static boost::uniform_int<> alluint(INT_MIN, INT_MAX);
static boost::variate_generator<boost::mt19937&, boost::uniform_int<> > 
	tr1mt_value(tr1mt, alluint);
#else
u32 tr1mt_value(void) { return 0; }
#endif

urandom::urandom(RandType method, u32 seed) {
	type = method;
	if (seed)
		seedvalue = seed;
	else
		seedvalue = time(NULL) ^ clock();

	if (type == Rand_Urandom)
		rnd.open("/dev/urandom", ios::in | ios::binary);
	else if (type == Rand_System)
		srand(seedvalue);
	else if (type == Rand_Cran32)
		cran32.Reset(time(NULL), clock());
	else if (type == Rand_Lcg)
		lcg.seed(seedvalue);
	else if (type == Rand_Rtprng)
		srtprng(seedvalue);
	else if (type == Rand_Isaac)
		initisaac();
	else if (type == Rand_Sfmt)
		init_gen_rand(seedvalue);
#ifdef USE_BOOST
	else if (type == Rand_Tr1mt)
		tr1mt.seed(seedvalue);
#endif

	totalused = 0;
	totalacquired = 0;
}

urandom::~urandom() {
	if (type == Rand_Urandom)
		rnd.close();
}
void urandom::settype(RandType method, u32 seed) {
	if (type == Rand_Urandom)
		rnd.close();

	type = method;
	if (seed)
		seedvalue = seed;
	else
		seedvalue = time(NULL) ^ clock();

	if (type == Rand_Urandom)
		rnd.open("/dev/urandom", ios::in | ios::binary);
	else if (type == Rand_System)
		srand(seedvalue);
	else if (type == Rand_Cran32)
		cran32.Reset(time(NULL), clock());
	else if (type == Rand_Lcg)
		lcg.seed(seedvalue);
	else if (type == Rand_Rtprng)
		srtprng(seedvalue);
	else if (type == Rand_Isaac)
		initisaac();
	else if (type == Rand_Sfmt)
		init_gen_rand(seedvalue);
#ifdef USE_BOOST
	else if (type == Rand_Tr1mt)
		tr1mt.seed(seedvalue);
#endif
}

u32 urandom::getused() {
	return totalused;
}

u32 urandom::getacquired() {
	return totalacquired;
}

u32 urandom::getseed() {
	return seedvalue;
}

u32 urandom::get(u8 len) { // in bits, up to 32
	u32 randbyte = 0;

	if (type == Rand_Sfmt) {
		totalused += len/8;
		totalacquired += 4;
		randbyte = gen_rand32();
	} else if (type == Rand_System) {
		totalused += len/8;
		totalacquired += 4;
		randbyte = rand();
	} else if (type == Rand_Cran32) {
		totalused += len/8;
		totalacquired += 4;
		randbyte = cran32.Next();
	} else if (type == Rand_Lcg) {
		totalused += len/8;
		totalacquired += 4;
		randbyte = lcg.nextNumber();
	} else if (type == Rand_Rtprng) {
		totalused += len/8;
		totalacquired += 4;
		randbyte = rtprng();
	} else if (type == Rand_Isaac) {
		totalused += len/8;
		totalacquired += 4;
		randbyte = getisaac();
	} else if (type == Rand_Urandom) {
		totalused += len/8;
		totalacquired += ((len + 7)/8);
		rnd.read((char *)&randbyte, (len + 7)/8);
	} else if (type == Rand_Tr1mt) {
		totalused += len/8;
		totalacquired += 4;
		randbyte = tr1mt_value();
	}

	if (len >= 32)
		return randbyte;
	return randbyte & ((0x1 << len) - 1);
}

u32 urandom::get32() { // in bits, up to 32
	u32 randbyte = 0;
	totalused += 4;
	totalacquired += 4;

	if (type == Rand_Sfmt)
		randbyte = gen_rand32();
	else if (type == Rand_System)
		randbyte = rand();
	else if (type == Rand_Cran32)
		randbyte = cran32.Next();
	else if (type == Rand_Lcg)
		randbyte = lcg.nextNumber();
	else if (type == Rand_Rtprng)
		randbyte = rtprng();
	else if (type == Rand_Isaac)
		randbyte = getisaac();
	else if (type == Rand_Urandom)
		rnd.read((char *)&randbyte, 4);
	else if (type == Rand_Tr1mt)
		randbyte = tr1mt_value();

	return randbyte;
}

float urandom::getfloat() {
	u32 randval = get32();
	// now scale to float and return
	float output = (float)randval / (float)UINT_MAX;
	return output;
}

// Implements Box-Muller transformation to get Gaussian distribution from uniform distribution
float urandom::getnormal(float mean, float stddev) {
	float x1, x2, w, y1;
	static float y2;
	static int use_last = 0;

	if (use_last) { // saves time
		y1 = y2;
		use_last = 0;
	} else {
		do {
			x1 = 2.0 * getfloat() - 1.0;
			x2 = 2.0 * getfloat() - 1.0;
			w = x1 * x1 + x2 * x2;
		} while ( w >= 1.0 );

		w = sqrt( (-2.0 * log( w ) ) / w );
		y1 = x1 * w;
		y2 = x2 * w;
		use_last = 1;
	}

	return( mean + y1 * stddev );
}

s32 urandom::get_inrange(s32 min, s32 max) {
	s32 randbyte;

	randbyte = get32() % (u32)(max - min + 1);

	return min + randbyte;
}
